"""Data classes for :mod:`auraxium.ps2.zone_effect`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'ZoneEffectData',
    'ZoneEffectTypeData'
]


class ZoneEffectData(Ps2Data):
    """Data class for :class:`auraxium.ps2.zone_effect.ZoneEffectData`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        zone_effect_id: The unique ID of this zone effect.
        zone_effect_type_id: The ID of the associated
            :class:`ZoneEffectType`.
        ability_id: The :class:`~auraxium.ps2.Ability` associated with
            this zone effect.
        param*: Type-specific parameters for this zone effect. Refer to
            the corresponding :class:`ZoneEffectType` for details.

    """

    zone_effect_id: int
    zone_effect_type_id: int
    ability_id: int
    param1: Optional[str] = None
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None
    param6: Optional[str] = None


class ZoneEffectTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.zone_effect.ZoneEffectType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        zone_effect_type_id: The unique ID of this zone effect type.
        description: A description of what this zone effect type is
            used for.
        param*: Descriptions of what the corresponding parameter is
            used for in zone effects of this type.

    """

    zone_effect_type_id: int
    description: str
    param1: Optional[str]
    param2: Optional[str]
    param3: Optional[str]
    param4: Optional[str]
    param5: Optional[str]
    param6: Optional[str]



"""Custom exceptions specific to the auraxium module."""

from typing import Any, Dict, Optional

import aiohttp
import yarl

__all__ = [
    'AuraxiumException',
    'CensusError',
    'UnknownCollectionError',
    'ServiceUnavailableError',
    'InvalidServiceIDError',
    'MissingServiceIDError',
    'BadRequestSyntaxError',
    'ServerError',
    'InvalidSearchTermError',
    'MaintenanceError',
    'NotFoundError',
    'ResponseError',
    'PayloadError'
]


class AuraxiumException(Exception):
    """Base exception class for auraxium.

    This mostly exists to allow blanket-catching any errors specific to
    the auraxium module.
    """


class CensusError(AuraxiumException):
    """Raised for server-side API errors.

    These are errors that are reported by the API endpoint and mostly
    inform of syntax errors, as well as invalid operations or API
    outages.

    Attributes:
        url: The URL that gave rise to the error.

    """

    def __init__(self, message: str, url: yarl.URL) -> None:
        super().__init__(message)
        self.url = url


class UnknownCollectionError(CensusError):
    """Raised if the queried collection does not exist.

    For the purposes of this error, a collection is tied to its
    namespace (aka. game and platform). Together, they form a qualified
    collection name, e.g. ``ps2/character`` or ``dcuo/world``.
    """

    def __init__(self, message: str, url: yarl.URL, namespace: str,
                 collection: Optional[str]) -> None:
        super().__init__(message, url)
        self.namespace = namespace
        self.collection = collection


class ServiceUnavailableError(CensusError):
    """Raised if the API reports a service as temporarily unavailable.

    This is likely to occur after big patches, or when the API is
    undergoing maintenance or has other issues.

    The issue causing this error is often specific to individual
    collections like ``ps2/characters_online_status``, the rest of the
    API might be fine.

    Unlike :class:`MaintenanceError`, this error state often persists
    for hours or even days.
    """


class InvalidServiceIDError(CensusError):
    """Raised when using an unknown service ID.

    Make sure your service ID is spelled correctly.

    If you only just applied for a custom service ID, you will have to
    wait for an email response before it is valid and usable.
    """


class MissingServiceIDError(CensusError):
    """Raised when using the default service ID.

    This error is raised when using the default service ID,
    ``s:example``, beyond its 10 requests per minute limit.

    This is easily fixed by applying for a custom service ID at
    http://census.daybreakgames.com/#devSignup.

    Once you have your own service ID, pass it to the
    :class:`auraxium.Client` or :class:`auraxium.census.Query`
    class initialiser respectively.
    """


class BadRequestSyntaxError(CensusError):
    """Raised if the query string itself was erroneous.

    This error points to a bug in the :mod:`auraxium.census` module. If
    you encounter it, please submit an issue at
    https://github.com/leonhard-s/auraxium/issues.
    """


class ServerError(CensusError):
    """Server-side error encountered when parsing the query string.

    This error generally points to a bug either the
    :mod:`auraxium.census` URL generator, or the game-specific object
    model.

    If you encounter this instance in the wild (i.e. not one of its
    subclasses), please submit an issue at
    https://github.com/leonhard-s/auraxium/issues.
    """


class InvalidSearchTermError(ServerError):
    """Server error for invalid search term strings.

    This means received when misusing API collections, or when trying
    to perform an invalid or unsupported operation on a given field.

    Attributes:
        collection: The collection that was accessed.
        namespace: The namespace of the collection.
        field: The field that caused the error. Might be ``None`` if
            the culprit could not be determined.

    """

    def __init__(self, message: str, url: yarl.URL, namespace: str,
                 collection: str, field: Optional[str]) -> None:
        super().__init__(message, url)
        self.namespace = namespace
        self.collection = collection
        self.field = field


class MaintenanceError(CensusError):
    """Raised if the API is down or undergoing maintenance."""

    def __init__(self, message: str, url: yarl.URL,
                 response: Optional[aiohttp.ClientResponse]) -> None:
        super().__init__(message, url)
        self.response = response


class ResponseError(AuraxiumException):
    """Raised for unexpected or invalid API responses.

    This can generally not be handled. Please submit an issue at
    https://github.com/leonhard-s/auraxium/issues if you encounter this
    error repeatedly.
    """


class PayloadError(AuraxiumException):
    """Raised if the payload returned by the API is unexpected.

    This generally points to a bug in the object model, or to the API
    having changed and requiring the object model to be updated.

    Please submit an issue at
    https://github.com/leonhard-s/auraxium/issues either way.
    """

    def __init__(self, message: str, payload: Dict[str, Any]) -> None:
        super().__init__(message)
        self.payload = payload


class NotFoundError(AuraxiumException):
    """Raised if a request unexpectedly did not return any matches."""



"""Object representations for the PlanetSide 2 API.

This is a higher-level abstraction of the data accessible through the
census. Note that the object attributes and hierarchy of this object
model will not match up to the Census API perfectly.
"""

from .ability import Ability, AbilityType, ResourceType
from .achievement import Achievement
from .armour import ArmourFacing, ArmourInfo
from .character import Character, Title
from .currency import Currency
from .depot import MarketingBundle, MarketingBundleSingle
from .directive import (Directive, DirectiveTier, DirectiveTree,
                        DirectiveTreeCategory)
from .effect import Effect, EffectType, TargetType
from .experience import Experience, ExperienceRank
from .faction import Faction
from .fire import FireGroup, FireMode, FireModeType
from .item import Item, ItemCategory, ItemType
from . import leaderboard
from .map import MapHex, MapRegion, Region
from .metagame import MetagameEvent, MetagameEventState
from .objective import Objective, ObjectiveType
from .outfit import Outfit, OutfitMember
from .profile import Loadout, Profile
from .projectile import Projectile, ProjectileFlightType
from .resist import ResistInfo, ResistType
from .reward import Reward, RewardType
from .skill import Skill, SkillCategory, SkillLine, SkillSet
from .states import PlayerState, PlayerStateGroup
from .vehicle import Vehicle, VehicleAttachment
from .weapon import Weapon, WeaponAmmoSlot, WeaponDatasheet
from .world import World
from .zone import Zone
from .zone_effect import ZoneEffect, ZoneEffectType

__all__ = [
    'Ability',
    'AbilityType',
    'Achievement',
    'ArmourFacing',
    'ArmourInfo',
    'Character',
    'Currency',
    'MarketingBundle',
    'MarketingBundleSingle',
    'Directive',
    'DirectiveTier',
    'DirectiveTree',
    'DirectiveTreeCategory',
    'Effect',
    'EffectType',
    'Experience',
    'ExperienceRank',
    'Faction',
    'FireGroup',
    'FireMode',
    'FireModeType',
    'Item',
    'ItemCategory',
    'ItemType',
    'leaderboard',
    'Loadout',
    'MapHex',
    'MapRegion',
    'Region',
    'MetagameEvent',
    'MetagameEventState',
    'Objective',
    'ObjectiveType',
    'Outfit',
    'OutfitMember',
    'Profile',
    'Projectile',
    'ProjectileFlightType',
    'ResistInfo',
    'ResistType',
    'ResourceType',
    'Reward',
    'RewardType',
    'Skill',
    'SkillCategory',
    'SkillLine',
    'SkillSet',
    'TargetType',
    'Title',
    'PlayerState',
    'PlayerStateGroup',
    'Vehicle',
    'VehicleAttachment',
    'Weapon',
    'WeaponAmmoSlot',
    'WeaponDatasheet',
    'World',
    'Zone',
    'ZoneEffect',
    'ZoneEffectType'
]



"""URL generation and validation utility."""

import warnings
from typing import Dict, Iterable, List, Optional, Tuple, Union

import yarl

from .support import JoinedQueryData, QueryData

__all__ = [
    'generate_url',
    'process_join',
]

REST_ENDPOINT = 'https://census.daybreakgames.com'


def generate_url(query: QueryData, verb: str, validate: bool = True) -> yarl.URL:
    """Generate the URL for a given query.

    This will also recursively process any joined queries.

    Arguments:
        query: The top-level query to process.
        verb: The query verb to use for the query.
        validate (optional): By default, the URL generator will perform
            a number of checks to validate the query, raising errors or
            warnings as necessary. Disabling this flag will skip the
            checks. Defaults to ``True``.

    Returns:
        A yarl URL representing the query.

    """
    # NOTE: The yarl.URL object uses the division operator to chain URI
    # components.

    # Census endpoint
    url = yarl.URL(REST_ENDPOINT)
    # Service ID
    url /= query.service_id
    if validate and query.service_id == 's:example':
        warnings.warn('The default service ID is heavily rate-limited. '
                      'Consider applying for your own service ID at '
                      'https://census.daybreakgames.com/#devSignup')
    # Query verb
    url /= verb
    # Namespace
    url /= query.namespace
    # Collection
    if (collection := query.collection) is not None:
        url /= collection
    elif validate:
        if query.terms:
            warnings.warn(f'No collection specified, but {len(query.terms)} '
                          'query terms provided')
        elif query.joins:
            warnings.warn(f'No collection specified, but {len(query.joins)} '
                          'joined queries provided')
    # Top-level query terms
    url = url.with_query([t.as_tuple() for t in query.terms])
    # Process query commands
    url = url.update_query(process_query_commands(query, validate=validate))
    return url


def process_join(data: JoinedQueryData, verbose: bool) -> str:
    """Return a string representation of the joined query.

    This generates is the string that will be inserted into the
    URL. This will also recursively process any inner joins added.

    Arguments:
        data: The data representing the joined query.
        verbose(optional): By default, the serialisation will try
            to save space by omitting fields left at their default
            value. Set this flag to True to change that. Defaults
            to ``False``.

    Returns:
        The string representation of the joined query.

    """
    # The collection (sometimes referred to as "type" in the docs) to join
    string = 'type:' if verbose else ''
    string += str(data.collection)
    # The fields used to link the two collections
    if (parent := data.field_on) is not None:
        string += f'^on:{parent}'
    if (child := data.field_to) is not None:
        string += f'^to:{child}'
    # Flags
    if (is_list := data.is_list) or verbose:
        string += '^list:' + ('1' if is_list else '0')
    if not (is_outer := data.is_outer) or verbose:
        string += '^outer:' + ('1' if is_outer else '0')
    # Show/hide field lists
    if show := data.show:
        string += '^show:' + '\''.join(str(s) for s in show)
    elif hide := data.hide:
        string += '^hide:' + '\''.join(str(s) for s in hide)
    # Inject at name
    if (name := data.inject_at) is not None:
        string += f'^inject_at:{name}'
    # QueryBase terms
    if terms := data.terms:
        string += '^terms:' + '\''.join(t.serialise() for t in terms)
    # Process nested (inner) joins
    if joins := data.joins:
        string += f'({",".join(process_join(j, verbose) for j in joins)})'
    return string


def process_query_commands(data: QueryData,
                           validate: bool = True) -> Dict[str, str]:
    """Process any query commands defined for the given query.

    This also recursively processes any joins defined.

    Arguments:
        data: The top-level query to process.
        validate (optional): Whether to perform checks on the query and
            warn the user about bad arguments. Defaults to ``True``.

    Returns:
        A dict of all query commands for the given query, this will be
        an empty dict if the query does not use any query commands.

    """
    commands: Dict[str, str] = {}
    # c:show
    if show := data.show:
        commands['show'] = ','.join(str(f) for f in show)
        if validate and data.hide:
            warnings.warn('Query.show and Query.hide are mutually exclusive, '
                          'the latter will be ignored')
    # c:hide
    elif hide := data.hide:
        commands['hide'] = ','.join(str(f) for f in hide)
    # c:sort
    if sort := data.sort:
        commands['sort'] = ','.join(_process_sorts(sort))
    # c:has
    if has := data.has:
        commands['has'] = ','.join(has)
    # c:resolve
    if resolve := data.resolve:
        commands['resolve'] = ','.join(resolve)
    # c:case
    if not data.case:
        commands['case'] = '0'
    # c:limit
    if (limit := data.limit) > 1:
        commands['limit'] = str(limit)
        if validate and data.limit_per_db > 1:
            warnings.warn('Query.limit and Query.limit_per_db are mutually '
                          'exclusive, the latter will be ignored')
    # c:limitPerDB
    elif (limit_per_db := data.limit_per_db) > 1:
        commands['limitPerDB'] = str(limit_per_db)
    # c:start
    if (start := data.start) > 0:
        commands['start'] = str(start)
    # c:includeNull
    if data.include_null:
        commands['includeNull'] = '1'
    # c:lang
    if (lang := data.lang) is not None:
        commands['lang'] = lang
    # c:join
    if data.joins:
        commands['join'] = ','.join(
            process_join(j, False) for j in data.joins)
    # c:tree
    if (tree := data.tree) is not None:
        commands['tree'] = _process_tree(tree)
    # c:timing
    if data.timing:
        commands['timing'] = '1'
    # c:exactMatchFirst
    if data.exact_match_first:
        commands['exactMatchFirst'] = '1'
    # c:distinct
    if (distinct := data.distinct) is not None:
        commands['distinct'] = distinct
    # c:retry
    if not data.retry:
        commands['retry'] = '0'
    # Add the 'c:' prefix to all of the keys
    commands = {f'c:{k}': v for k, v in commands.items()}
    return commands


def _process_sorts(sorts: Iterable[Union[str, Tuple[str, bool]]]) -> List[str]:
    """Process a top-level query's sort_by attribute into a list.

    This mostly handles the sorting direction tuples.

    Arguments:
        sorts: The sorting values to process.

    Raises:
        ValueError: Raised if an invalid sorting key is encountered.

    Returns:
        A list of sorting fields with sort order tokens.

    """
    processed: List[str] = []
    for sort in sorts:
        # Plain strings can be kept as-is
        if isinstance(sort, str):
            processed.append(sort)
        # Non-string objects are expected to be tuples showing sort order
        else:
            try:
                field, sort_order = sort
            except ValueError as err:
                raise ValueError(f'Invalid sort key: {sort}') from err
            processed.append(field if sort_order else f'{field}:-1')
    return processed


def _process_tree(tree: Dict[str, Optional[Union[str, bool]]]) -> str:
    """Process the dict created by the :meth:`Query.as_tree()` method.

    Arguments:
        tree: The mapping to process.

    Returns:
        The string representation of the tree command.

    """
    string = str(tree['field'])
    if prefix := tree['prefix']:
        string += f'^prefix:{prefix}'
    if tree['is_list']:
        string += '^list:1'
    if (start := tree['start']) is not None:
        string += f'^start:{start}'
    return string



"""Data classes for :mod:`auraxium.ps2.currency`."""

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'CurrencyData'
]


class CurrencyData(Ps2Data):
    """Data class for :class:`auraxium.ps2.currency.Currency`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        currency_id: The unique ID of this currency entry.
        name: The localised name of this currency.
        icon_id: The image ID of the currency icon image asset.
        inventory_cap: The maximum amount of this currency a character
            may hold.

    """

    currency_id: int
    name: LocaleData
    icon_id: int
    inventory_cap: int



"""Request handling and response checking for Auraxium.

This module is responsible for performing the HTTP requests themselves,
handling any HTTP-related errors, and converting the returned data into
a dictionary that can be parsed by the object model.

It also checks for API error codes and raises the appropriate
exceptions.

All requests are buffered, allowing up to five HTTP-related errors
before giving up and letting the error propagate outwards.
"""

import json
import logging
import sys
import warnings
from typing import Any, Dict, List, Optional, Tuple

import aiohttp
import backoff
import yarl

from .census import Query
from .errors import (PayloadError, BadRequestSyntaxError, CensusError,
                     InvalidSearchTermError, InvalidServiceIDError,
                     MaintenanceError, MissingServiceIDError, NotFoundError,
                     ResponseError, ServerError, ServiceUnavailableError,
                     UnknownCollectionError)
from .types import CensusData
from .utils import expo_scaled

__all__ = [
    'extract_payload',
    'extract_single',
    'run_query'
]

log = logging.getLogger('auraxium.http')


def get_components(url: yarl.URL) -> Tuple[str, Optional[str]]:
    """Return the namespace and collection of a given query.

    Arguments:
        url: The :class:`yarl.URL` to process. Only REST API query URLs
            in the DBG census API format are allowed.

    Returns:
        The namespace/game and collection that was accessed. Collection
        may be ``None`` for some queries.

    """
    components = url.path.split('/')[1:]
    if components[0].startswith('s:'):  # Remove service ID
        _ = components.pop(0)
    if components[0] in ('xml', 'json'):  # Remove format specifier
        _ = components.pop(0)
    _ = components.pop(0)  # Remove query verb
    if not components[-1]:  # Remove collection if empty
        _ = components.pop(-1)
    # If only a namespace was provided, return None as the collection
    if len(components) == 1:
        return components[0], None
    assert len(components) == 2, 'Unable to parse URL'
    return components[0], components[1]


async def response_to_dict(response: aiohttp.ClientResponse) -> CensusData:
    """Convert a response received from the server to a dictionary.

    In some cases - mostly error states - the API will return JSON data
    without providing the appropriate ``Content-Type`` entity header,
    which breaks :meth:`aiohttp.ClientResponse.json()`.

    This function catches this error and generates the expected
    dictionary from the plain text response instead, if possible. If
    the native :mod:`json` module cannot parse the response either, a
    :class:`~auraxium.errors.ResponseError` is raised.

    Arguments:
        response: The :class:`aiohttp.ClientResponse` to convert.

    Raises:
        ResponseError: Raised if the response cannot be converted into
            a dictionary.

    Returns:
        A dictionary containing the response payload.

    """
    data: CensusData
    try:
        # NOTE: It is possible to skip content type checking completely by
        # setting the content_type flag to False and letting its JSON decoder
        # error out within aiohttp, then handle that - this is arguably neater.
        data = await response.json()
    except aiohttp.ContentTypeError as err:
        text = await response.text()
        # Run the plain text through the JSON decoder anyway before we let the
        # error propagate to handle the misrepresented content type issue.
        try:
            data = json.loads(text)
        except json.JSONDecodeError:
            # There really is something wrong with this data, let an updated
            # error propagate.
            raise ResponseError(
                f'Received a non-JSON response: {text}') from err
        log.info(
            'Received a plain text response containing JSON data: %s', text)
        return data
    return data


def extract_payload(data: CensusData, collection: str) -> List[CensusData]:
    """Extract the payload from a census response dictionary.

    Arguments:
        data: The response dictionary to process.
        collection: The collection name to expect.

    Raises:
        BadPayloadError: Raised if the provided response dictionary
            lacks the collection-specific key expected.

    Returns:
        All dictionaries in the response list.

    """
    try:
        list_: List[CensusData] = data[f'{collection}_list']
    except KeyError as err:
        raise PayloadError(
            f'Unable to extract list of results due to missing key '
            f'"{collection}_list" in payload: {data}', data) from err
    return list_


def extract_single(data: CensusData, collection: str,
                   no_warn_multi: bool = False) -> CensusData:
    """Extract the payload from a census response dictionary.

    Identical to :meth:`extract_payload`, but only returns the first
    match. If multiple results are found, a warning will be issued.

    Arguments:
        data: The response dictionary to process.
        collection: The collection name to expect.
        no_warn_multi (optional): Set this flag to disable the warning
            issued when using this function on a multi-result payload.

    Raises:
        BadPayloadError: Raised if the provided response dictionary
            lacks the collection-specific key expected.
        NotFoundError: Raised if the collection does not contain any
            results.

    Returns:
        The first result payload in the given dictionary.

    """
    try:
        list_ = extract_payload(data, collection)
    except PayloadError as err:
        raise PayloadError(
            f'Unable to extract result due to missing key "{collection}_list" '
            f'in payload: {data}', data) from err
    if not list_:
        raise NotFoundError('The server did not return any matches')
    if not no_warn_multi and len(list_) > 1:
        warnings.warn('The dictionary passed to extract_single contained '
                      'multiple results. Check your query limit settings or '
                      'use extract_payload if you expect a list of results.')
    return list_[0]


def raise_for_dict(data: CensusData, url: yarl.URL) -> None:
    """Check an API response dictionary for errors.

    This raises the appropriate :class:`auraxium.errors.CensusError`
    subclass for a given API response.

    Arguments:
        data: The API response dictionary to check.
        url: The URL used to perform the request. This is included in
            any errors raised to facilitate troubleshooting.

    Raises:
        UnknownCollectionError: Raised when attempting to query a
            namespace or collection that does not exist.
        BadRequestSyntaxError: Raised if the server reports an
            erroneous query.
        ServiceUnavailableError: Raised if the API service accessed is
            currently unavailable.
        InvalidServiceIDError: Raised when using invalid service IDs.
        MissingServiceIDError: Raised when exceeding the rate limits
            imposed on the public ``s:examples`` service ID.
        ServerError: Raised for server-side query parsing errors.
        InvalidSearchTermError: Raised for bad field names or values
        CensusError: Raised for unknown server-side errors.

    """
    # Syntax parser and namespace dispatching
    if (msg := data.get('error')) is not None:
        if msg == 'No data found.':
            namespace, collection = get_components(url)
            # NOTE: This error is returned if either the namespace or
            # collection are unknown
            if collection is None:
                raise UnknownCollectionError(
                    f'The namespace "{namespace}" does not exist.',
                    url, namespace, collection)
            help_url = Query(namespace=namespace)
            raise UnknownCollectionError(
                f'No collection at "{namespace}/{collection}", try {help_url} '
                f'for a the list of valid collections, or an error message if '
                f'the the namespace "{namespace}" is invalid.',
                url, namespace, collection)
        if msg == 'Bad request syntax.':
            raise BadRequestSyntaxError(
                'An error occurred in the auraxium.census URL generator. '
                'Please report this error at '
                'https://github.com/leonhard-s/auraxium/issues', url)
        if msg == 'service_unavailable':
            raise ServiceUnavailableError(
                'This component of the API is currently unavailable. '
                'Try again later.', url)
        if msg.startswith('Provided Service ID is not registered.'):
            raise InvalidServiceIDError(
                'The service ID specified is unknown. Check your spelling '
                'or wait for the email response validating your service ID.',
                url)
        if msg.startswith('Missing Service ID.'):
            raise MissingServiceIDError(
                'The default service ID "s:example" is for casual use only. '
                'Wait 60 seconds or apply for your own service ID at '
                'http://census.daybreakgames.com/#devSignup', url)
    # Collection field and query command parsing
    if (code := data.get('errorCode')) is not None:
        if (msg := data.get('errorMessage')) is not None:
            if code == 'SERVER_ERROR':
                if msg.startswith('INVALID_SEARCH_TERM'):
                    _process_invalid_search_term(msg, url)
                raise ServerError(
                    f'Error code: "{code}", error message: "{msg}"', url)
        if code == 'SERVER_ERROR':
            # NOTE: There are at least two circumstances under which there is
            # no message associated with a server error:
            # 1. An empty request (i.e. https://census.daybreakgames.com/get/)
            # 2. Not providing any arguments to ps2/event
            raise ServerError('Unknown server error', url)
        # Fallback for new and exciting error codes
        raise CensusError(
            f'An unknown error code was encountered: {code}', url)


def _process_invalid_search_term(msg: str, url: yarl.URL) -> None:
    """Parse "INVALID_SEARCH_TERM" error messages.

    This reformats the error message and tries to return a more helpful
    error message through URL introspection.

    This method will return if it cannot provide a specific error.

    Arguments:
        msg: The error message received.
        url: The URL that resulted in the current error.

    Raises:
        InvalidSearchTermError: Raised for bad field names or values

    """
    # Process the URL to provide more helpful errors
    *_, namespace, collection = url.parts
    # Shorter version of the message, with "INVALID_SEARCH_TERM: " chopped off
    chopped = msg[21:].strip()
    # Invalid field name
    if chopped.startswith('Invalid search term. Valid search terms:'):
        # Retrieve the list of valid field names from the error message
        fields_str = chopped.split(':', maxsplit=1)[1].strip()
        fields: List[str] = [f.strip() for f in fields_str[1:-1].split(',')]
        # Parse the query string to find the faulty field name
        culprit: Optional[str] = None
        for field, value in url.query.items():
            if field not in fields:
                culprit = field
                break
        # Tweak the error message depending on whether a culprit was found
        field_info = 'field' if culprit is None else f'"{culprit}"'
        raise InvalidSearchTermError(
            f'Invalid {field_info} for collection '
            f'"{namespace}/{collection}". Valid field names: {fields}',
            url, namespace, collection, culprit)
    # Invalid field value
    if chopped.startswith('Invalid search term:'):
        field = chopped.split(':', maxsplit=1)[1].split()[0][:-1]
        value = url.query[field]
        raise InvalidSearchTermError(
            f'Invalid value "{value}" for field "{field}" of collection '
            f'"{namespace}/{collection}".', url, namespace, collection, field)
    # Invalid RegEx value
    if chopped.startswith('Invalid search value for term:'):
        field = chopped.split(':', maxsplit=1)[1].split()[0][:-1]
        value = url.query[field]
        raise InvalidSearchTermError(
            f'Invalid value "{value}" for field "{field}" of collection '
            f'"{namespace}/{collection}". At least three characters must be '
            'provided in addition to the RegEx flag.',
            url, namespace, collection, field)
    # No valid terms in show/hide fields
    if chopped.startswith('c:show or c:hide resulted'):
        method = 'show' if 'c:show' in str(url) else 'hide'
        raise InvalidSearchTermError(
            f'Invalid field names specified for QueryBase.{method}().',
            url, namespace, collection, f'c:{method}')


async def run_query(query: Query, session: aiohttp.ClientSession,
                    verb: str = 'get') -> CensusData:
    """Perform a top-level Query using the provided HTTP session.

    This will handle check both the HTTP response and JSON contents for
    errors before returning.

    Refer to the :meth:`raise_for_dict()` method for details on the
    exceptions raised.

    Arguments:
        query: The query to run.
        session: The session to use for the request.
        verb (optional): The query verb to pass. Defaults to ``'get'``.

    Raises:
        ResponseError: Raised if the HTTP response contained error
            codes or could not be parsed.

    Returns:
        The response dictionary received.

    """
    url = query.url(verb=verb)
    log.debug('Performing %s request: %s', verb.upper(), url)

    def on_success(details: Dict[str, Any]) -> None:
        """Called when a query is successful."""
        if (tries := details['tries']) > 1:
            log.debug('Query successful after %d tries: %s',
                      tries, url)

    def on_backoff(details: Dict[str, Any]) -> None:
        """Called when a query failed and is backed off."""
        wait = details['wait']
        tries = details['tries']
        log.debug('Backing off %.2f seconds after %d attempts: %s',
                  wait, tries, url)

    def on_giveup(details: Dict[str, Any]) -> None:
        """Called when giving up on a query."""
        elapsed: float = details['elapsed']
        tries: int = details['tries']
        log.warning('Giving up on query and re-raising exception after %.2f '
                    'seconds and %d attempts: %s', elapsed, tries, url)
        _, exc_value, _ = sys.exc_info()
        assert exc_value is not None
        raise exc_value

    # The following exceptions will be retried if occurring during the request
    backoff_errors = (
        aiohttp.ClientResponseError,
        aiohttp.ClientConnectionError,
        MaintenanceError)

    @backoff.on_exception(expo_scaled(0.05), backoff_errors,  # type: ignore
                          on_backoff=on_backoff, on_giveup=on_giveup,
                          on_success=on_success, max_time=30.0, max_tries=10,
                          jitter=None)
    async def retry_query() -> aiohttp.ClientResponse:
        """Request handling wrapper."""
        response = await session.get(
            url, allow_redirects=False, raise_for_status=True)
        # Trigger MaintenanceErrors from redirect response. This will also be
        # caught by the backoff decorator and will only reach the user if the
        # logic in the on_backoff callback re-raises it.
        if 300 <= response.status < 400:
            raise MaintenanceError(
                'API redirection detected, API maintenance inferred',
                url, response)
        return response

    # Check for HTTP errors
    try:
        response = await retry_query()
    except aiohttp.ClientResponseError as err:
        raise ResponseError(
            f'An HTTP exception occurred: {err.args[0]}') from err
    except aiohttp.ClientConnectionError as err:
        # The connection had issues.
        raise ResponseError(
            f'A network exception occurred: {err.args[0]}') from err

    # Convert the HTTP response into a dictionary
    data = await response_to_dict(response)
    # Check the received dictionary for error codes
    raise_for_dict(data, url)
    return data



"""Data classes for :mod:`auraxium.ps2.profile`."""

from ..base import Ps2Data

__all__ = [
    'LoadoutData',
    'ProfileData'
]


class LoadoutData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.Loadout`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        loadout_id: The unique ID of this loadout.
        profile_id: The ID of the associated profile.
        faction_id: The faction for this loadout.
        code_name: A string describing the loadout.

    """

    loadout_id: int
    profile_id: int
    faction_id: int
    code_name: str


class ProfileData(Ps2Data):
    """Data class for :class:`auraxium.ps2.profile.Profile`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        profile_id: The unique ID of this profile.
        description: The description of the profile.

    """

    profile_id: int
    description: str



"""Data classes for :mod:`auraxium.ps2.skill`."""

from typing import Optional

from ..base import ImageData, Ps2Data
from ..types import LocaleData


__all__ = [
    'SkillData',
    'SkillCategoryData',
    'SkillLineData',
    'SkillSetData'
]


class SkillData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.skill.Skill`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        skill_id: The unique ID of this skill.
        skill_line_id: The ID of the associated :class:`SkillLine`.
        skill_line_index: The position of the skill in its skill line.
        skill_points: The unlock cost of the skill.
        grant_item_id: The :class:`~auraxium.ps2.Item` granted by this
            skill, if any.
        name: The localised name of this skill.
        description: The localised description for this skill.

    """

    skill_id: int
    skill_line_id: int
    skill_line_index: int
    skill_points: int
    grant_item_id: Optional[int] = None
    name: LocaleData
    description: Optional[LocaleData] = None


class SkillCategoryData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.skill.SkillCategory`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        skill_category_id: The unique ID of this skill category.
        skill_set_id: The :class:`SkillCategory` this category belongs
            to.
        skill_set_index: The position of this category in the
            associated skill category.
        skill_points: The unlock cost for this skill category.
        name: The localised name of this skill category.
        description: The localised description for this skill category.

    """

    skill_category_id: int
    skill_set_id: int
    skill_set_index: int
    skill_points: int
    name: LocaleData
    description: Optional[LocaleData] = None


class SkillLineData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.skill.SkillLine`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        skill_line_id: The unique ID for this skill line.
        skill_points: The unlock cost for this skill line.
        skill_category_id: The :class:`SkillCategory` this skill line
            belongs to.
        skill_category_index: The index of this skill line in its
            containing skill category.
        name: The localised name of this skill line.
        description: The localised description for this skill line.

    """

    skill_line_id: int
    skill_points: int
    skill_category_id: Optional[int] = None
    skill_category_index: Optional[int] = None
    name: LocaleData
    description: Optional[LocaleData] = None


class SkillSetData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.skill.SkillSet`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        skill_set_id: The unique ID of this skill set.
        skill_points: (Not yet documented)
        required_item_id: The item required to unlock this skill set.
            Used to prevent buying upgrades for items the player has
            not unlocked yet.
        name: The localised name of the skill set.
        description: The localised description of the skill set.

    """

    skill_set_id: int
    skill_points: Optional[int] = None
    required_item_id: Optional[int] = None
    name: LocaleData
    description: Optional[LocaleData] = None



"""Ability and ability type class definitions."""

from ..base import Cached
from ..census import Query
from ..models import AbilityData, AbilityTypeData, ResourceTypeData
from ..proxy import InstanceProxy


class ResourceType(Cached, cache_size=50, cache_ttu=3600.0):
    """A type of resource consumed by an ability."""

    collection = 'resource_type'
    data: ResourceTypeData
    dataclass = ResourceTypeData
    id_field = 'resource_type_id'

    @property
    def description(self) -> str:
        """A description of what this resource type is used for."""
        return self.data.description


class AbilityType(Cached, cache_size=20, cache_ttu=60.0):
    """A type of ability.

    This class mostly specifies the purpose of any generic parameters.
    """

    collection = 'ability_type'
    data: AbilityTypeData
    dataclass = AbilityTypeData
    id_field = 'ability_type_id'


class Ability(Cached, cache_size=10, cache_ttu=60.0):
    """An ability cast by a character.

    Access the corresponding :class:`auraxium.ps2.ability.AbilityType`
    instance via the :meth:`type` method for information on generic
    parameters.
    """

    collection = 'ability'
    data: AbilityData
    dataclass = AbilityData
    id_field = 'ability_id'

    def resource_type(self) -> InstanceProxy[ResourceType]:
        """Return the resource type used by this ability, if any."""
        type_id = self.data.resource_type_id or -1
        query = Query(
            ResourceType.collection, service_id=self._client.service_id)
        query.add_term(field=ResourceType.id_field, value=type_id)
        return InstanceProxy(ResourceType, query, client=self._client)

    def type(self) -> InstanceProxy[AbilityType]:
        """Return the ability type of this ability."""
        query = Query(
            AbilityType.collection, service_id=self._client.service_id)
        query.add_term(
            field=AbilityType.id_field, value=self.data.ability_type_id)
        return InstanceProxy(AbilityType, query, client=self._client)



"""Vehicle class definitions."""

from typing import Final, List, Optional, Union

from ..base import Cached, ImageMixin, Named
from ..census import Query
from ..client import Client
from ..models import VehicleAttachmentData, VehicleData
from ..proxy import InstanceProxy, SequenceProxy

from .faction import Faction
from .item import Item
from .skill import SkillSet


class Vehicle(Named, ImageMixin, cache_size=50, cache_ttu=3600.0):
    """A mountable vehicle in PlanetSide 2.

    This includes aircraft and ground vehicles, as well as mountable
    turrets and constructible.
    """

    collection = 'vehicle'
    data: VehicleData
    dataclass = VehicleData
    id_field = 'vehicle_id'

    def factions(self) -> SequenceProxy[Faction]:
        """Return the factions that have access to this vehicle.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'vehicle_faction'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=Vehicle.id_field, value=self.id)
        query.limit(5)
        return SequenceProxy(Faction, query, client=self._client)

    @classmethod
    async def get_by_faction(cls, faction: Union[Faction, int], *,
                             client: Client) -> List['Vehicle']:
        """Return all vehicles available to the given faction."""
        collection: Final[str] = 'vehicle_faction'
        faction_id = faction.id if isinstance(faction, Faction) else faction
        query = Query(collection, service_id=client.service_id)
        query.add_term(field=Faction.id_field, value=faction_id)
        query.limit(500)
        join = query.create_join(cls.collection)
        join.set_fields(cls.id_field)
        proxy: SequenceProxy['Vehicle'] = SequenceProxy(
            cls, query, client=client)
        return await proxy.flatten()

    def skill_sets(self, faction: Optional[Union[Faction, int]] = None
                   ) -> SequenceProxy[SkillSet]:
        """Return the skill sets associated with this vehicle.

        By default, this will list all skill sets for all empires. Note
        that this may result in duplicate results as some skill sets
        exist for all empires.
        To avoid duplicates, either generate a set from the returned
        proxy, or specify the faction to resolve skill sets for.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'vehicle_skill_set'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        if faction is not None:
            faction = faction.id if isinstance(faction, Faction) else faction
            query.add_term(field=Faction.id_field, value=faction)
        query.limit(500)
        query.sort('display_index')
        join = query.create_join(SkillSet.collection)
        join.set_fields(SkillSet.id_field)
        return SequenceProxy(SkillSet, query, client=self._client)


class VehicleAttachment(Cached, cache_size=250, cache_ttu=180.0):
    """Links vehicles to the items and attachments they support."""

    collection = 'vehicle_attachment'
    data: VehicleAttachmentData
    dataclass = VehicleAttachmentData
    id_field = 'vehicle_attachment_id'

    def faction(self) -> InstanceProxy[Faction]:
        """Return the faction this attachment is available to.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Faction.collection, service_id=self._client.service_id)
        query.add_term(field=Faction.id_field, value=self.data.faction_id)
        return InstanceProxy(Faction, query, client=self._client)

    def item(self) -> InstanceProxy[Item]:
        """Return the attachable item for the vehicle.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Item.collection, service_id=self._client.service_id)
        query.add_term(field=Item.id_field, value=self.data.item_id)
        return InstanceProxy(Item, query, client=self._client)

    def vehicle(self) -> InstanceProxy[Vehicle]:
        """Return the vehicle the item may be attached to.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Vehicle.collection, service_id=self._client.service_id)
        query.add_term(field=Vehicle.id_field, value=self.data.vehicle_id)
        return InstanceProxy(Vehicle, query, client=self._client)



"""Data classes for :mod:`auraxium.ps2.outfit`."""

from ..base import Ps2Data

__all__ = [
    'OutfitData',
    'OutfitMemberData',
    'OutfitRankData'
]


class OutfitData(Ps2Data):
    """Data class for :class:`auraxium.ps2.outfit.Outfit`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        outfit_id: The unique ID of the outfit.
        name: The name of the outfit.
        name_lower: Lowercase version of :attr`name`. Useful for
            optimising case-insensitive searches.
        alias: The alias (or tag) of the outfit.
        alias_lower: Lowercase version of :attr:`alias`. Useful for
            optimising case-insensitive searches.
        time_created: The creation date of the outfit as a UTC
            timestamp.
        time_created_date: Human-readable version of
            :attr:`time_created`.
        leader_character_id: The character/member ID of the outfit
            leader.
        member_count: The number of members in the outfit.

    """

    outfit_id: int
    name: str
    name_lower: str
    alias: str
    alias_lower: str
    time_created: int
    time_created_date: str
    leader_character_id: int
    member_count: int


class OutfitMemberData(Ps2Data):
    """Data class for :class:`auraxium.ps2.outfit.OutfitMember`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        outfit_id: The ID of the outfit this member is a part of.
        character_id: The ID of the associated character.
        member_since: The date the character joined the outfit at as
            a UTC timestamp.
        member_since_date: Human-readable version of
            :attr:`member_since`.
        rank: The name of the member's in-game outfit rank.
        rank_ordinal: The ordinal position of the member's rank within
            the outfit. The lower the value, the higher the rank.

    """

    outfit_id: int
    character_id: int
    member_since: int
    member_since_date: str
    rank: str
    rank_ordinal: int


class OutfitRankData(Ps2Data):
    """Data class for :class:`auraxium.ps2.outfit.OutfitRank`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        outfit_id: The unique ID of the outfit the rank belongs to.
        ordinal: The position of the rank within the outfit, lower
            values indicate higher ranks.
        name: The name of the rank.
        description: The description of the rank.

    """

    outfit_id: int
    ordinal: int
    name: str
    description: str



"""Shared utility methods used throughout Auraxium."""

from typing import Callable, Iterator

import backoff

__all__ = [
    'expo_scaled'
]


def expo_scaled(factor: float = 1.0, max_: float = 3600.0
                ) -> Callable[[], Iterator[float]]:
    """Reusable factory for scaling :meth:`backoff.expo` values.

    This allows reducing the initial retry times for request retries.
    The returned function can be passed to the regular backoff handlers
    like :meth:`backoff.on_exception`.

    Arguments:
        factor (optional): The scaling factor for the exponential.
        Defaults to ``1.0``.
        max_ (optional): The maximum delay to wait for. Defaults to
        ``3600.0``.

    Returns:
        A callable that returns an iterator returning the delay to use
        in seconds.

    """
    gen: Iterator[float] = backoff.expo(base=2, max_value=max_)  # type: ignore

    def inner() -> Iterator[float]:
        """A scaled version :meth:`backoff.expo`."""
        while True:
            try:
                yield next(gen) * factor
            except StopIteration:
                return

    return inner



"""Alert and alert state class definitions."""

import enum

from ..base import Cached
from ..models import MetagameEventData


class MetagameEventState(enum.IntEnum):
    """The state of a :class:`auraxium.ps2.metagame.MetagameEvent`."""

    STARTED = 135
    RESTARTED = 136
    CANCELLED = 137
    ENDED = 138
    XP_BONUS_CHANGED = 139


class MetagameEvent(Cached, cache_size=100, cache_ttu=60.0):
    """An event or alert on a continent."""

    collection = 'metagame_event'
    data: MetagameEventData
    dataclass = MetagameEventData
    id_field = 'metagame_event_id'



"""Custom types used by the auraxium module."""

from typing import Any, Callable, Dict, Optional, TypeVar

import pydantic

__all__ = [
    'CensusData',
    'LocaleData',
    'optional'
]

T = TypeVar('T')  # pylint: disable=invalid-name


# NOTE: The inner dict's value is typed as "Any" due to Mypy not supporting
# cycling definitions yet.
CensusData = Dict[str, Any]


class LocaleData(pydantic.BaseModel):
    """Container for localised strings.

    Note that the ``tr`` locale is ignored as it was abandoned by the
    developers and is generally either missing or unpopulated.
    """

    class Config:
        """Pydantic model configuration.

        This inner class is used to namespace the pydantic
        configuration options.
        """
        allow_mutation = False

    de: Optional[str] = None
    en: Optional[str] = None
    es: Optional[str] = None
    fr: Optional[str] = None
    it: Optional[str] = None

    @classmethod
    def empty(cls) -> 'LocaleData':
        """Return an empty :class:`LocaleData` instance.

        This is mostly provided to easily handle payloads who's entire
        localised string field is ``NULL``.
        """
        return cls()


def optional(data: CensusData, key: str,
             cast: Callable[[Any], T]) -> Optional[T]:
    """Cast an optional dictionary value to a given type.

    This is a helper method that acts much like :meth:`dict.get()`, but
    also casts the retrieved value to the given type if it is not None.

    Arguments:
        data: The dictionary to process.
        key: The key to access.
        cast: The type to cast the value to if it exists.

    Returns:
        The cast value retrieved from the dictionary, or None.

    """
    raw: Optional[T]
    if (raw := data.pop(key, None)) is not None:
        if raw == 'NULL':
            raw = None
        else:
            raw = cast(raw)
    return raw



"""Utility functions for working with leaderboard data."""

import enum
import warnings
from typing import Dict, Final, List, Optional, Tuple, Union

from ..census import Query
from ..errors import NotFoundError
from ..client import Client
from ..request import extract_payload, extract_single

from .character import Character
from .world import World

__all__ = ['by_char', 'by_char_multi', 'Period', 'Stat', 'top']


class Period(enum.Enum):
    """The periods supported by the leaderboard."""

    FOREVER = 0
    MONTHLY = 1
    WEEKLY = 2
    DAILY = 3
    ONE_LIFE = 4


class Stat(enum.Enum):
    """A statistic tracked by the leaderboard."""

    KILLS = 0
    SCORE = 1
    TIME = 2
    DEATHS = 3


async def by_char(stat: Stat, character: Union[int, Character],
                  period: Period = Period.FOREVER,
                  *, client: Client) -> Optional[Tuple[int, int]]:
    """Return the rank of the player on the leaderboard.

    Note that only the top 10'000 players are tracked by the
    leaderboard. For characters not in the top 10'000, this will return
    None. Otherwise, this returns the a tuple consisting of the rank
    and stat value for the character.
    """
    char_id = character if isinstance(character, int) else character.id
    collection: Final[str] = 'characters_leaderboard'
    query = Query(collection, service_id=client.service_id)
    query.add_term(field=Character.id_field, value=char_id)
    query.add_term(field='name', value=_name_from_stat(stat))
    query.add_term(field='period', value=_period_from_enum(period))
    payload = await client.request(query)
    try:
        data = extract_single(payload, collection)
    except NotFoundError:
        return None
    return int(data['rank']), int(data['value'])


async def by_char_multi(stat: Stat, character: Union[int, Character],
                        *args: Union[int, Character],
                        period: Period = Period.FOREVER,
                        client: Client) -> List[Tuple[int, int]]:
    """Return the rank of the players on the leaderboard.

    Like by_char, but takes any number of arguments.
    """
    char_ids = [character if isinstance(character, int) else character.id]
    char_ids.extend([c if isinstance(c, int) else c.id for c in args])
    value = ','.join(str(c) for c in char_ids)
    collection: Final[str] = 'characters_leaderboard'
    query = Query(collection, service_id=client.service_id)
    query.add_term(field=Character.id_field, value=value)
    query.add_term(field='name', value=_name_from_stat(stat))
    query.add_term(field='period', value=_period_from_enum(period))
    payload = await client.request(query)
    data = extract_payload(payload, collection)
    return_: Dict[int, Tuple[int, int]] = {i: (-1, -1) for i in char_ids}
    for row in data:
        id_ = int(row['character_id'])
        return_[id_] = int(row['rank']), int(row['value'])
    return [return_[i] for i in char_ids]


async def top(stat: Stat, period: Period = Period.FOREVER, matches: int = 10,
              offset: int = 0, world: Optional[Union[int, World]] = None,
              *, client: Client) -> List[Tuple[int, Character]]:
    """Retrieve the top entries on the leaderboard for the given stat.

    Note that only the top 10'000 players are tracked by the
    leaderboard.
    """
    if matches > 10_000:
        warnings.warn('Results will been truncated to 10k elements due to '
                      'API limitations')
        matches = 10_000
    collection: Final[str] = 'leaderboard'
    query = Query(collection, service_id=client.service_id)
    query.add_term(field='name', value=_name_from_stat(stat))
    query.add_term(field='period', value=_period_from_enum(period))
    if world is not None:
        world_id: int = world if isinstance(world, int) else world.id
        query.add_term(field='world', value=world_id)
    query.limit(matches)
    query.start(offset)
    join = query.create_join(Character.collection)
    join.set_fields(Character.id_field)
    payload = await client.request(query)
    key = 'character_id_join_character'
    return [(int(d['value']), Character(d[key], client=client))
            for d in extract_payload(payload, collection=collection)]


def _name_from_stat(stat: Stat) -> str:
    periods = ['Kills', 'Score', 'Time', 'Deaths']
    return periods[stat.value]


def _period_from_enum(period: Period) -> str:
    names = ['Forever', 'Monthly', 'Weekly', 'Daily', 'OneLife']
    return names[period.value]



"""Weapon class definition."""

import logging
from typing import Final, List, Optional

from ..base import Cached
from ..client import Client
from ..census import Query
from ..models import WeaponAmmoSlot, WeaponData, WeaponDatasheet
from ..proxy import InstanceProxy, SequenceProxy
from ..request import extract_payload, extract_single

from .fire import FireGroup
from .item import Item

__all__ = [
    'Weapon',
    'WeaponAmmoSlot',
    'WeaponDatasheet'
]

log = logging.getLogger('auraxium')


class Weapon(Cached, cache_size=128, cache_ttu=3600.0):
    """A weapon available to a player.

    This can be treated as an extension to the
    :class:`auraxium.ps2.item.Item` class.
    """

    collection = 'weapon'
    data: WeaponData
    dataclass = WeaponData
    id_field = 'weapon_id'

    @property
    def is_heat_weapon(self) -> bool:
        """Guess whether this weapon is using a heat mechanic.

        This checks for presence and non-zero value for the
        "heat_mechanic" stat.
        """
        if (capacity := self.data.heat_capacity) is not None:
            return capacity > 0
        return False

    async def ammo_slots(self) -> List[WeaponAmmoSlot]:
        """Return the ammo slots for the weapon."""
        collection: Final[str] = 'weapon_ammo_slot'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(10).sort('weapon_slot_index')
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        return [WeaponAmmoSlot(**d) for d in data]

    def attachments(self) -> SequenceProxy[Item]:
        """Return the attachments available for this weapon.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'weapon_to_attachment'
        group_id = self.data.weapon_group_id or -1
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field='weapon_group_id', value=group_id)
        query.limit(100)
        join = query.create_join(Item.collection)
        join.set_fields(Item.id_field)
        join.set_outer(False)
        return SequenceProxy(Item, query, client=self._client)

    async def datasheet(self) -> WeaponDatasheet:
        """Return the datasheet for the weapon."""
        collection: Final[str] = 'weapon_datasheet'
        if (item := await self.item()) is None:
            raise RuntimeError(f'Invalid item for weapon ID: {self.id}')
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=Item.id_field, value=item.id)
        payload = await self._client.request(query)
        data = extract_single(payload, collection)
        return WeaponDatasheet(**data)

    def fire_groups(self) -> SequenceProxy[FireGroup]:
        """Return the fire groups for this weapon.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'weapon_to_fire_group'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(20)
        join = query.create_join(FireGroup.collection)
        join.set_fields(FireGroup.id_field)
        return SequenceProxy(FireGroup, query, client=self._client)

    @classmethod
    async def get_by_name(cls, name: str, *, locale: str = 'en',
                          client: Client) -> Optional['Weapon']:
        """Retrieve a weapon by name.

        This is a helper method provided as weapons themselves do not
        have a name. This looks up an item by name, then returns the
        weapon associated with this item.

        Returns:
            The weapon associated with the given item, or None

        """
        item = await Item.get_by_name(name, locale=locale, client=client)
        if item is None:
            return None
        return await item.weapon().resolve()

    def item(self) -> InstanceProxy[Item]:
        """Return the item associated with this weapon."""
        query = Query('item_to_weapon', service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        join = query.create_join('item')
        join.set_fields('item_id', None)
        return InstanceProxy(Item, query, client=self._client)



"""Bundles and special offer class definitions."""

from typing import Final, List, Tuple

from ..base import Named, Cached
from ..census import Query
from ..models import MarketingBundleData, MarketingBundleSingleData
from ..proxy import InstanceProxy
from ..request import extract_payload

from .item import Item


class MarketingBundle(Named, cache_size=100, cache_ttu=60.0):
    """A marketing bundle containing one or more items.

    This is used for special promotions, or for bundles that contain
    multiple items at once.
    """

    collection = 'marketing_bundle'
    data: MarketingBundleData
    dataclass = MarketingBundleData
    id_field = 'marketing_bundle_id'

    def image(self) -> str:
        """Return the default image for this type."""
        image_id: int = self.data.image_id  # type: ignore
        url = 'https://census.daybreakgames.com/files/ps2/images/static/'
        return url + f'{image_id}.png'

    async def items(self) -> List[Tuple[Item, int]]:
        """Return the contents of the bundle.

        This returns a list of tuples consisting of the item, followed
        by the quantity.
        """
        collection: Final[str] = 'marketing_bundle_item'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(100)
        join = query.create_join(Item.collection)
        join.set_fields(Item.id_field)
        payload = await self._client.request(query)
        data = extract_payload(payload, collection)
        key_name = f'{Item.id_field}_join_{Item.collection}'
        items: List[Tuple[Item, int]] = []
        for item_data in data:
            item = Item(item_data[key_name], client=self._client)
            count = int(item_data['quantity'])
            items.append((item, count))
        return items


class MarketingBundleSingle(Cached, cache_size=100, cache_ttu=60.0):
    """A marketing bundle containing a single item.

    This is used for single-item entries in the depot, such as weapons,
    scopes or other items that do not require any additional
    information.
    """

    collection = 'marketing_bundle_with_1_item'
    data: MarketingBundleSingleData
    dataclass = MarketingBundleSingleData
    id_field = 'marketing_bundle_id'

    def item(self) -> InstanceProxy[Item]:
        """Return the item unlocked by the bundle.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Item.collection, service_id=self._client.service_id)
        query.add_term(field=Item.id_field, value=self.data.item_id)
        return InstanceProxy(Item, query, client=self._client)



"""Query objects definition and supporting code.

This module defines the interfaces provided by top-level queries and
inner, nested queries respectively.
"""

import copy
from typing import Any, List, Optional, Tuple, Type, TypeVar, Union

import yarl

from .support import (CensusValue, JoinedQueryData, QueryBaseData, QueryData,
                      SearchModifier, SearchTerm)
from .urlgen import generate_url

__all__ = [
    'JoinedQuery',
    'Query',
    'QueryBase'
]

_QueryBaseT = TypeVar('_QueryBaseT', bound='QueryBase')
_T = TypeVar('_T')


class QueryBase:
    """Base class for functionality shared between queries and joins.

    If you want to re-use the same query multiple times, use this
    class. For more control about how the query is performed, refer to
    the :class:`Query` and :class:`JoinedQuery` subclasses.

    To convert between different types of queries, use the
    :meth:`QueryBase.copy()` factory. See that method's docstring for
    details.

    Attributes:
        data: Provides low-level access to the represented query
        joins: A list of inner queries that were attached to this one.

    """

    def __init__(self, collection: Optional[str] = None,
                 **kwargs: CensusValue) -> None:
        """Initialise the query.

        Arguments:
            collection (optional): The API collection to access or
                ``None`` to display the list of available collections.
                Defaults to ``None``.
            *kwargs: Key/value pairs to pass to the
                :meth:`QueryBase.add_term()` method.

        """
        self.data = QueryBaseData(collection)
        self.joins: List[JoinedQuery] = []
        # Replace and double underscores with dots to allow accessing inner
        # fields like "name.first" or "battle_rank.value"
        kwargs = {k.replace('__', '.'): v for k, v in kwargs.items()}
        # Run the add_term method for each of the converted key/value pairs
        _ = [self.add_term(k, v, parse_modifier=True)
             for k, v in kwargs.items()]

    def add_join(self: _QueryBaseT, query: 'JoinedQuery',
                 **kwargs: Any) -> _QueryBaseT:
        """Add an existing :class:`JoinedQuery` to this query.

        This converts an existing :class:`QueryBase` instance to a
        :class:`JoinedQuery` using the :meth:`QueryBase.copy()`
        factory. The created join is then added to this query.

        To create a new query and add it immediately, use the
        :meth:`QueryBase.create_join` method instead.

        Arguments:
            query: Another query to join to the current query.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.joins.append(JoinedQuery.copy(query, **kwargs))
        return self

    def add_term(self: _QueryBaseT, field: str, value: CensusValue,
                 modifier: SearchModifier = SearchModifier.EQUAL_TO, *,
                 parse_modifier: bool = False) -> _QueryBaseT:
        """Add a new filter term to the query.

        Filter terms are used to either reduce the number of results
        returned, or to specify the exact ID expected. Refer to the
        :class:`SearchTerm` class for details and examples.

        Arguments:
            field: The field to filter by.
            value: The value of the filter term.
            modifier(optional): A search modifier to use. This will
                only be used if ``parse_modifier`` is False.
                Defaults to ``SearchModifier.EQUAL_TO``.
            parse_modifier(optional): If True, the search modifier
                will be inferred from the value. Defaults to ``False``.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        if parse_modifier:
            term = SearchTerm.infer(field, value)
        else:
            term = SearchTerm(field, value, modifier=modifier)
        self.data.terms.append(term)
        return self

    @classmethod
    def copy(cls: Type[_QueryBaseT], template: 'QueryBase',
             copy_joins: bool = False, deep_copy: bool = False,
             **kwargs: Any) -> _QueryBaseT:
        """Create a new query, copying most data from the template.

        The new query will share the collection, terms and show/hide
        markers of the template. If ``copy_joins`` is enabled, it will
        also copy its list of joins.

        Among other things, allows easy creation of joins from existing
        queries, which is handy if you have complex existing joins or
        hidden fields that would be tedious to recreate.

        By default, this creates a shallow copy. Modifying the terms or
        joined queries will cause mutations of the template. Set the
        ``deep_copy`` flag to ensure complete independence.

        Any keyword arguments are passed to the new query's
        initialiser.

        Example:

            .. code-block:: python3

                # This is an existing query that does what we need it
                # to, assume it has some complex join or hidden field
                # setup that would make it tedious to re-create.
                old = Query('character')

                # This is an unrelated, new query. We want its join to
                # return the exact same data structure as the previous
                # query.
                new = Query('outfit_member', outfit_id=...).limit(1000)

                # Create a join emulating the original query and add it
                join = JoinedQuery.copy(old, copy_joins=True)
                new.add_join(join)

        Arguments:
            template: The query to copy.
            copy_joins (optional): Whether to recursively copy joined
                queries. Defaults to ``False``.
            deep_copy (optional): Whether to perform a deep copy. Use
                this if you intend to modify the list of terms or other
                mutable types to avoid changing the template. Defaults
                to ``False``.
            **kwargs: Any keyword arguments are passed on to the new
                query's constructor.

        Raises:
            TypeError: Raised when attempting to copy into a
                :class:`JoinedQuery` without a collection specified.

        Returns:
            An instance of the current class populated with information
            from the template query.

        """
        copy_func = copy.deepcopy if deep_copy else _dummy_copy
        # Create a new querybase instance
        instance = cls(copy_func(template.data.collection), **kwargs)
        attrs = ['terms', 'hide', 'show']
        for attr in attrs:
            value = copy_func(getattr(template.data, attr))
            setattr(instance.data, attr, value)
        if copy_joins:
            instance.joins = copy_func(template.joins)
        return instance

    def create_join(self, collection: str, *args: Any,
                    **kwargs: Any) -> 'JoinedQuery':
        """Create a new joined query and add it to the current one.

        See the initialiser for JoinedQuery for arguments, this method
        passes and parameters given on.

        Arguments:
            collection: The collection to join.
            *args: Any anonymous positional arguments are passed on to
                :meth:`JoinedQuery.__init__()`
            *kwargs: Any anonymous keyword arguments are passed on to
                :meth:`JoinedQuery.__init__()`

        Returns:
            The JoinedQuery instance that was created.

        """
        join = JoinedQuery(collection, *args, **kwargs)
        self.joins.append(join)
        return join

    def hide(self: _QueryBaseT, field: str, *args: str) -> _QueryBaseT:
        """Set the fields to hide in the response.

        The given fields will not be included in the result. Note that
        this can break joins if the field they are attached to is not
        included.

        This is mutually exclusive with :meth:`QueryBase.show()`;
        setting one will undo any changes made by the other.

        Arguments:
            field: A field name to hide from the result data.
            *args: Any number of additional fields to hide.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.hide = [field]
        self.data.hide.extend(args)
        self.data.show = []
        return self

    def show(self: _QueryBaseT, field: str, *args: str) -> _QueryBaseT:
        """Set the fields to show in the response.

        Any other fields will not be included in the result. Note that
        this can break joins if the field they are attached to is not
        included.

        This is mutually exclusive with :meth:`QueryBase.hide()`;
        setting one will undo any changes made by the other.

        Arguments:
            field: A field name to include in the result data.
            *args: Any number of additional fields to include.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.show = [field]
        self.data.show.extend(args)
        self.data.hide = []
        return self


class Query(QueryBase):
    """The main query supplied to the API.

    The top-level query has access to additional return value formats
    such as sorting or tree views, and also supports additional, global
    flags that will propagate through to any inner, joined queries.

    This subclasses :class:`QueryBase`. Refer to its docstring for
    details on inherited methods and attributes.

    You can find additional information on the attributes in their
    respective setter methods.

    Attributes:
        data: Provides low-level access to the represented query

    """

    def __init__(self, collection: Optional[str] = None,
                 namespace: str = 'ps2:v2', service_id: str = 's:example',
                 **kwargs: CensusValue) -> None:
        """Create a new top-level query.

        The collection argument, as well as any keyword arguments, are
        passed on to the constructor of the QueryBase class.

        Arguments:
            collection (optional): The API collection to access.
                Defaults to ``None``.
            namespace (optional): The namespace to access. Defaults to
                ``'ps2:v2'``.
            service_id (optional): Your personal service ID. Note that
                the default service ID is heavily rate limited.
                Defaults to ``'s:example'``.

        """
        super().__init__(collection, **kwargs)
        data: QueryBaseData = self.data  # type: ignore
        self.data = QueryData.from_base(data)
        self.data.namespace = namespace
        self.data.service_id = service_id

    def __str__(self) -> str:
        """Return the string representation of the query.

        This is the URL in its finished form. Use :meth:`Query.url()`
        to retrieve the URL as a :class:`yarl.URL` instance for extra
        control.

        Returns:
            The full URL describing this query and all of its joins.

        """
        return str(self.url().human_repr())

    def case(self, value: bool = True) -> 'Query':
        """Globally ignore case for this query.

        Note that case-insensitive look-ups are significantly slower.
        Where available, use a case-sensitive query targeting a
        lowercase field like ``ps2/character.name.first_lower``.

        Arguments:
            value (optional): Whether to ignore case for this query.
                Defaults to ``True``.

        Returns:
            The full URL describing this query and all of its joins.

        """
        self.data.case = value
        return self

    @classmethod
    def copy(cls, template: 'QueryBase', copy_joins: bool = False,
             deep_copy: bool = False, **kwargs: Any) -> 'Query':
        """Create a new query, copying most data from the template.

        The new query will share the collection, terms and show/hide
        markers of the template. If ``copy_joins`` is enabled, it will
        also copy its list of joins.

        Among other things, allows easy creation of joins from existing
        queries, which is handy if you have complex existing joins or
        hidden fields that would be tedious to recreate.

        By default, this creates a shallow copy. Modifying the terms or
        joined queries will cause mutations of the template. Set the
        ``deep_copy`` flag to ensure complete independence.

        Any keyword arguments are passed to the new query's
        initialiser.

        Example:

            .. code-block:: python3

                # This is an existing query that does what we need it
                # to, assume it has some complex join or hidden field
                # setup that would make it tedious to re-create.
                old = Query('character')

                # This is an unrelated, new query. We want its join to
                # return the exact same data structure as the previous
                # query.
                new = Query('outfit_member', outfit_id=...).limit(1000)

                # Create a join emulating the original query and add it
                join = JoinedQuery.copy(old, copy_joins=True)
                new.add_join(join)

        Arguments:
            template: The query to copy.
            copy_joins (optional): Whether to recursively copy joined
                queries. Defaults to ``False``.
            deep_copy (optional): Whether to perform a deep copy. Use
                this if you intend to modify the list of terms or other
                mutable types to avoid changing the template. Defaults
                to ``False``.
            **kwargs: Any keyword arguments are passed on to the new
                query's constructor.

        Raises:
            TypeError: Raised when attempting to copy into a
                :class:`JoinedQuery` without a collection specified.

        Returns:
            An instance of the current class populated with information
            from the template query.

        """
        copy_func = copy.deepcopy if deep_copy else _dummy_copy
        # Create a new Query instance
        instance: 'Query' = super().copy(  # type: ignore
            template, copy_joins=copy_joins, deep_copy=deep_copy, *kwargs)
        if isinstance(template, Query):
            # Additional attributes to include when copying a top-level query
            attrs = ['case', 'distinct', 'exact_match_first', 'has',
                     'include_null', 'lang', 'limit', 'limit_per_db',
                     'namespace', 'resolve', 'retry', 'service_id', 'sort',
                     'start', 'timing', 'tree']
            for attr in attrs:
                value = copy_func(getattr(template.data, attr))
                setattr(instance.data, attr, value)
        # Include an arbitrary limit when copying from a joined list
        elif isinstance(template, JoinedQuery) and template.data.is_list:
            # NOTE: Joined lists have no set length, so this might break for
            # very long or complex joins.
            instance.limit(10000)  # pylint: disable=no-member
        return instance

    def has(self, field: str, *args: str) -> 'Query':
        """Hide results with a ``NULL`` value at the given field.

        This is useful for filtering large data sets by optional
        fields, such as searching the ``ps2/weapons`` collection for
        heat-based weapons using the heat-mechanic-specific fields.

        Arguments:
            field: The field required for results to be included.
            *args: Additional required fields.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.has = [field]
        self.data.has.extend(args)
        return self

    def distinct(self, field: Optional[str]) -> 'Query':
        """Query command used to show all unique values for a field.

        Arguments:
            field: The field to show unique values for. Set to ``None``
                to disable.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.distinct = field
        return self

    def exact_match_first(self, value: bool = True) -> 'Query':
        """Whether to display exact matches before partial matches.

        When performing RegEx searches (i.e. ones using either
        :attr:`SearchModifier.STARTS_WITH` or
        :attr:`SearchModifier.CONTAINS`), this setting will always
        promote an exact match to be the first item returned,
        regardless of any sorting settings applied.

        Arguments:
            value (optional): Whether to promote exact matches.
                Defaults to ``True``.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.exact_match_first = value
        return self

    def include_null(self, value: bool) -> 'Query':
        """Whether to include NULL values in the response.

        This is useful for API introspection, but it is generally more
        bandwidth-friendly to use the :meth:`dict.get()` method with a
        default value when parsing the result dictionary.

        This only affects the top-level query itself; joined queries
        will only show non-NULL values.

        Arguments:
            value: Enable or disable the setting.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.include_null = value
        return self

    def lang(self, lang: Optional[str] = None) -> 'Query':
        """Set the locale to user for the query.

        By default, queries return all locales for localised strings.
        Use this flag to only include the given locale, or reset to
        ``None`` to include all localisations.

        The following locales are currently supported and maintained:::

            German: 'de', English: 'en', Spanish: 'es',
            French: 'fr', Italian: 'it'

        Arguments:
            lang (optional): The locale identifier to return. Defaults
                to ``None``.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.lang = lang
        return self

    def limit(self, limit: int) -> 'Query':
        """Specify the number of results returned.

        By default, the API will only return the first match for any
        given query, you can increase the number of results using this
        method.

        The maximum number of values permissible varies from collection
        to collection, e.g. 100k for ``ps2/character``, but only 5000
        for ``ps2/item``. Use your best judgement.

        This is mutually exclusive with :meth:`Query.limit_per_db()`,
        setting one will undo the changes made by the other.

        Arguments:
            limit: The number of results to return. Must be at least 1.

        Raises:
            ValueError: Raised if ``limit`` is less than 1.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        if limit < 1:
            raise ValueError('limit must be greater than or equal to 1')
        self.data.limit = limit
        self.data.limit_per_db = 1
        return self

    def limit_per_db(self, limit_per_db: int) -> 'Query':
        """Specify the number of results returned per database.

        This method works similarly to :meth:`Query.limit()`, but will
        yield better results for distributed collections such as
        ps2/character, which is spread across 20 different databases
        more or less randomly.

        This is mutually exclusive with :meth:`Query.limit()`, setting
        one will undo the changes made by the other.

        Arguments:
            limit_per_db: The number of results to return per database.
                Must be at least 1.

        Raises:
            ValueError: Raised if ``limit_per_db`` is less than 1.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        if limit_per_db < 1:
            raise ValueError('limit_per_db must be greater than or equal to 1')
        self.data.limit_per_db = limit_per_db
        self.data.limit = 1
        return self

    def offset(self, offset: int) -> 'Query':
        """Alias for the :meth:`Query.start()` method.

        Refer to its docstring for details.

        Arguments:
            offset: The number of results to skip.

        Raises:
            ValueError: Raised if offset is negative.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        try:
            self.start(offset)
        except ValueError as err:
            raise ValueError('offset may not be negative') from err
        return self

    def resolve(self, name: str, *args: str) -> 'Query':
        """Resolve additional data for a collection.

        Resolves are a lighter version of joined queries and can be
        used to quickly include associated information with the
        results.

        Perform a query with no collection specified to see a list of
        resolvable names for each collection.

        Arguments:
            name: A resolvable name to attach to the query.
            *args: Any number of additional resolvable names to attach.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.resolve = [name]
        self.data.resolve.extend(args)
        return self

    def retry(self, retry: bool = False) -> 'Query':
        """Enable automatic query retry.

        By default, failed queries will be retried automatically. Set
        this to False to disable this behaviour if you want to fail
        early.

        Arguments:
            retry (optional): Whether to enable automatic query
                retrying. Defaults to ``False``.

        Returns:
            The full URL describing this query and all of its joins.

        """
        self.data.retry = retry
        return self

    def start(self, start: int) -> 'Query':
        """Skip the given number of results in the response.

        Together with :meth:`Query.limit()`, this can be used to create
        a paginated view of API data.

        Arguments:
            start: The number of results to skip.

        Raises:
            ValueError: Raised if start is negative.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        if start < 0:
            raise ValueError('start may not be negative')
        if start >= 0:
            self.data.start = start
        return self

    def sort(self, field: Union[str, Tuple[str, bool]],
             *args: Union[str, Tuple[str, bool]]) -> 'Query':
        """Sort the results by one or more fields.

        By default, this uses ascending sort order. For descending
        order, pass a tuple with a negative second element:

        .. code-block:: python3

            QueryBase.sort('field1', ('field'2, True))  # Ascending
            QueryBase.sort(('field3', False))  # Descending

        If multiple field names are provided, multiple sorting passes
        will be performed in order to further refine the list returned.

        Arguments:
            field: A qualified field name to sort by.
            *args: Additional fields to sort by.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.sort = [field]
        self.data.sort.extend(args)
        return self

    def timing(self, value: bool = True) -> 'Query':
        """Enabling query profiling output.

        Setting this flag will include an additional "timing" key in
        the response, providing timing information for the main query
        and any joins.

        Arguments:
            value (optional): Whether to enable profiling. Defaults to
                ``True``.

        Returns:
            The full URL describing this query and all of its joins.

        """
        self.data.timing = value
        return self

    def tree(self, field: str, is_list: bool = False, prefix: str = '',
             start: Optional[str] = None) -> 'Query':
        """Reformat a result list into a data tree.

        This is useful for lists of data with obvious categorisation,
        such as grouping weapons by their type.

        Arguments:
            field: The field to remove and use for the data structure.
            list (optional): Whether the tree data is a list. Defaults
                to 0.
            prefix (optional): A prefix to add to the field value to
                increase readability. Defaults to ``''``.
            start (optional): Used to tell the tree where to start. If
                ``None``, the root list of results will be reformatted
                as a tree. Defaults to ``None``.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.tree = {'field': field, 'is_list': is_list,
                          'prefix': prefix, 'start': start}
        return self

    def url(self, verb: str = 'get', skip_checks: bool = False) -> yarl.URL:
        """Generate the URL representing this query.

        This will also recursively process any joined queries.

        Arguments:
            verb (optional): The query verb to use for the query. Known
                options are ``'get'``, used to return a list of
                results, and ``'count'``, used to return the length of
                that list. Defaults to ``'get'``.
            skip_checks (optional): By default, the URL generator will
                perform a number of checks to validate your query.
                Enabling this flag will skip the checks. Defaults to
                ``False``.

        Returns:
            A URL object representing the query.

        """
        self.data.joins = [j.serialise() for j in self.joins]
        return generate_url(self.data, verb, validate=not skip_checks)


class JoinedQuery(QueryBase):
    """A sub-query to be joined to an existing query.

    Joined queries (or "joins") allow performing multiple, related
    look-ups in the same request. For a simpler but less powerful
    interface, see the :meth:`Query.resolve()` method.

    This subclasses :class:`QueryBase`. Refer to its docstring for
    details on inherited methods and attributes.

    You can find additional information on the attributes in their
    respective setter methods.

    Attributes:
        data: Provides low-level access to the represented joined query

    """

    def __init__(self, collection: str, **kwargs: CensusValue) -> None:
        """Instantiate a joined, inner query."""
        super().__init__(collection, **kwargs)
        data: QueryBaseData = self.data  # type: ignore
        self.data = JoinedQueryData.from_base(data)

    @classmethod
    def copy(cls, template: QueryBase, copy_joins: bool = False,
             deep_copy: bool = False, **kwargs: Any) -> 'JoinedQuery':
        """Create a new query, copying most data from the template.

        The new query will share the collection, terms and show/hide
        markers of the template. If ``copy_joins`` is enabled, it will
        also copy its list of joins.

        Among other things, allows easy creation of joins from existing
        queries, which is handy if you have complex existing joins or
        hidden fields that would be tedious to recreate.

        By default, this creates a shallow copy. Modifying the terms or
        joined queries will cause mutations of the template. Set the
        ``deep_copy`` flag to ensure complete independence.

        Any keyword arguments are passed to the new query's
        initialiser.

        Example:

            .. code-block:: python3

                # This is an existing query that does what we need it
                # to, assume it has some complex join or hidden field
                # setup that would make it tedious to re-create.
                old = Query('character')

                # This is an unrelated, new query. We want its join to
                # return the exact same data structure as the previous
                # query.
                new = Query('outfit_member', outfit_id=...).limit(1000)

                # Create a join emulating the original query and add it
                join = JoinedQuery.copy(old, copy_joins=True)
                new.add_join(join)

        Arguments:
            template: The query to copy.
            copy_joins (optional): Whether to recursively copy joined
                queries. Defaults to ``False``.
            deep_copy (optional): Whether to perform a deep copy. Use
                this if you intend to modify the list of terms or other
                mutable types to avoid changing the template. Defaults
                to ``False``.
            **kwargs: Any keyword arguments are passed on to the new
                query's constructor.

        Raises:
            TypeError: Raised when attempting to copy into a
                :class:`JoinedQuery` without a collection specified.

        Returns:
            An instance of the current class populated with information
            from the template query.

        """
        # A joined query cannot be created without a collection
        if template.data.collection is None:
            raise TypeError('JoinedQuery requires a collection')
        copy_func = copy.deepcopy if deep_copy else _dummy_copy
        # Create a new JoinedQuery instance
        instance: JoinedQuery = super().copy(  # type: ignore
            template, copy_joins=copy_joins, deep_copy=deep_copy, **kwargs)
        if isinstance(template, JoinedQuery):
            # Additional attributes to include when copying another join
            attrs = [
                'inject_at', 'is_list', 'is_outer', 'field_on', 'field_to']
            for attr in attrs:
                value = copy_func(getattr(template.data, attr))
                setattr(instance.data, attr, value)
        # If the original query had a non-default limit value, the join should
        # also return a list.
        elif isinstance(template, Query):
            if template.data.limit > 1 or template.data.limit_per_db > 1:
                instance.data.is_list = True
        return instance

    def serialise(self) -> JoinedQueryData:
        """Process any internal joins and return a nested data dict."""
        data = self.data
        data.joins = [j.serialise() for j in self.joins]
        return data

    def set_fields(self, parent: Optional[str], child: Optional[str] = None
                   ) -> 'JoinedQuery':
        """Set the field names to use for the join.

        The API will use inferred names whenever possible, inferred
        names might be ``<parent-collection>_id`` or`
        ``<child-collection>_id``.

        Use this method to specify the field names manually. Either of
        the given values may be ``None`` to use the default naming
        system.

        Specifying only the parent's name will apply it to both fields.

        Arguments:
            parent: The field name on the parent collection.
            child (optional): The field name on the child collection.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        if parent is not None:
            self.data.field_on = parent
            if child is None:
                self.data.field_to = parent
        if child is not None:
            self.data.field_to = child
        return self

    def set_inject_at(self, key: Optional[str]) -> 'JoinedQuery':
        """Set the name of the field to insert the joined data at.

        By default, the inserted is added to a dynamically generated
        key following the pattern
        ``<parent_field>_join_<joined_collection>``. Example:::

            character_id_join_characters_online_status

        This method allows overriding the name of this insertion key.

        This will overwrite existing keys. If the existing key is a
        JSON object/Python dict, the added keys will be merged. When
        updating this dictionary, any colliding keys will be appended
        with ``_merged``.
        Non-dict keys will simply be overwritten by the joined data.

        Arguments:
            key (optional): The name of the key to inject the joined
                data at. If ``None``, the autogenerated name is used.
                Defaults to ``None``.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.inject_at = key
        return self

    def set_list(self, is_list: bool) -> 'JoinedQuery':
        """Set whether the current join should return a list.

        If True, the join will return any matching elements. Be wary of
        large relational collections such as ``ps2/characters_item``;
        there is no limiting system, just an eventual hard cut-off. Use
        terms to reduce the number of matching elements when flagging a
        join as a list.

        Arguments:
            is_list: Whether the join should return a list.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.is_list = is_list
        return self

    def set_outer(self, is_outer: bool) -> 'JoinedQuery':
        """Set whether the current join is an outer or inner join.

        An outer join (the default setting) will include all results,
        regardless of whether some of the terms in its joins are met or
        not.

        An inner join will exclude these settings, which can be useful
        when filtering by inner values.

        For example, say you were displaying a ``ps2/characters_item``
        list with the associated ``ps2/item`` collection joined. Even
        if you added a term to only add the joins for items that are
        weapons, you would still find the full ``ps2/character_item``
        list in your results. This is the outer join behaviour.
        However, if the join is flagged as an inner join, it will
        discard any results that do not meet the join's terms.

        Arguments:
            is_outer: If True, the join will be an outer join. Set to
                False for inner join behaviour.

        Returns:
            The query instance; this allows for chaining of operations.

        """
        self.data.is_outer = is_outer
        return self


def _dummy_copy(obj: _T) -> _T:
    """Dummy function that does not actually copy anything."""
    return obj



"""Profile and loadout class definitions."""

from typing import Final

from ..base import Cached, FallbackMixin
from ..census import Query
from ..models import LoadoutData, ProfileData
from ..proxy import InstanceProxy, SequenceProxy
from ..types import CensusData

from .armour import ArmourInfo
from .faction import Faction
from .resist import ResistInfo


class Profile(Cached, cache_size=200, cache_ttu=60.0):
    """An entity in the game world.

    This is used to specify the resistance and armour values to
    apply to a given object.

    Profiles include faction-specific classes, vehicles, facility
    infrastructure such as turrets, generators or shields, as well as
    other non-static entities such as Cortium nodes or pumpkins.
    """

    collection = 'profile_2'
    data: ProfileData
    dataclass = ProfileData
    id_field = 'profile_id'

    def armour_info(self) -> SequenceProxy[ArmourInfo]:
        """Return the armour info of the profile.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'profile_armor_map'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(20)
        join = query.create_join(ArmourInfo.collection)
        join.set_fields(ArmourInfo.id_field)
        return SequenceProxy(ArmourInfo, query, client=self._client)

    def resist_info(self) -> SequenceProxy[ResistInfo]:
        """Return the resist info of the profile.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'profile_resist_map'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=self.id_field, value=self.id)
        query.limit(500)
        join = query.create_join(ResistInfo.collection)
        join.set_fields(ResistInfo.id_field)
        return SequenceProxy(ResistInfo, query, client=self._client)


def _get_fallback(id_: int) -> CensusData:
    profile_id = id_ + 162 if id_ != 45 else 252
    code_name = {
        28: 'NSO Infiltrator',
        29: 'NSO Light Assault',
        30: 'NSO Medic',
        31: 'NSO Engineer',
        32: 'NSO Heavy Assault',
        45: 'NSO MAX'}
    return {
        'loadout_id': id_,
        'profile_id': profile_id,
        'faction_id': 4,
        'code_name': code_name[id_]}


class Loadout(Cached, FallbackMixin, cache_size=20, cache_ttu=3600.0):
    """Represents a faction-specific infantry class."""

    collection = 'loadout'
    data: LoadoutData
    dataclass = LoadoutData
    id_field = 'loadout_id'
    _fallback = {k: _get_fallback(k) for k in (*range(28, 33), 45)}

    def armour_info(self) -> SequenceProxy[ArmourInfo]:
        """Return the armour info of the loadout.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'profile_armor_map'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=Profile.id_field, value=self.data.profile_id)
        query.limit(20)
        join = query.create_join(ArmourInfo.collection)
        join.set_fields(ArmourInfo.id_field)
        return SequenceProxy(ArmourInfo, query, client=self._client)

    def faction(self) -> InstanceProxy[Faction]:
        """Return the faction of the loadout.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Faction.collection, service_id=self._client.service_id)
        query.add_term(field=Faction.id_field, value=self.data.faction_id)
        return InstanceProxy(Faction, query, client=self._client)

    def profile(self) -> InstanceProxy[Profile]:
        """Return the profile of the loadout.

        This returns an :class:`auraxium.proxy.InstanceProxy`.
        """
        query = Query(Profile.collection, service_id=self._client.service_id)
        query.add_term(field=Profile.id_field, value=self.data.profile_id)
        return InstanceProxy(Profile, query, client=self._client)

    def resist_info(self) -> SequenceProxy[ResistInfo]:
        """Return the resist info of the loadout.

        This returns a :class:`auraxium.proxy.SequenceProxy`.
        """
        collection: Final[str] = 'profile_resist_map'
        query = Query(collection, service_id=self._client.service_id)
        query.add_term(field=Profile.id_field, value=self.data.profile_id)
        query.limit(500)
        join = query.create_join(ResistInfo.collection)
        join.set_fields(ResistInfo.id_field)
        return SequenceProxy(ResistInfo, query, client=self._client)



"""Python representations of PlanetSide 2 payloads."""

from .ability import AbilityData, AbilityTypeData, ResourceTypeData
from .achievement import AchievementData
from .armour import ArmourInfoData
from .character import (CharacterAchievement, CharacterData,
                        CharacterDirective, TitleData)
from .currency import CurrencyData
from .depot import MarketingBundleData, MarketingBundleSingleData
from .directive import (DirectiveData, DirectiveTierData,
                        DirectiveTreeCategoryData, DirectiveTreeData)
from .effect import EffectData, EffectTypeData
from .experience import ExperienceData, ExperienceRankData
from .faction import FactionData
from .fire import FireGroupData, FireModeData
from .item import ItemCategoryData, ItemData, ItemTypeData
from .map import FacilityTypeData, MapHexData, MapRegionData, RegionData
from .metagame import MetagameEventData
from .objective import ObjectiveData, ObjectiveTypeData
from .outfit import OutfitData, OutfitMemberData, OutfitRankData
from .profile import LoadoutData, ProfileData
from .projectile import ProjectileData
from .resist import ResistInfoData, ResistTypeData
from .reward import RewardData, RewardTypeData
from .skill import SkillData, SkillCategoryData, SkillLineData, SkillSetData
from .states import PlayerStateGroup
from .vehicle import VehicleAttachmentData, VehicleData
from .weapon import WeaponAmmoSlot, WeaponData, WeaponDatasheet
from .world import WorldData
from .zone import ZoneData
from .zone_effect import ZoneEffectData, ZoneEffectTypeData

__all__ = [
    'AbilityData',
    'AbilityTypeData',
    'AchievementData',
    'ArmourInfoData',
    'CharacterAchievement',
    'CharacterData',
    'CharacterDirective',
    'CurrencyData',
    'DirectiveData',
    'DirectiveTierData',
    'DirectiveTreeCategoryData',
    'DirectiveTreeData',
    'EffectData',
    'EffectTypeData',
    'ExperienceData',
    'ExperienceRankData',
    'FacilityTypeData',
    'FactionData',
    'FireGroupData',
    'FireModeData',
    'ItemCategoryData',
    'ItemData',
    'ItemTypeData',
    'LoadoutData',
    'MapHexData',
    'MapRegionData',
    'MarketingBundleData',
    'MarketingBundleSingleData',
    'MetagameEventData',
    'ObjectiveData',
    'ObjectiveTypeData',
    'OutfitData',
    'OutfitMemberData',
    'OutfitRankData',
    'PlayerStateGroup',
    'ProfileData',
    'ProjectileData',
    'RegionData',
    'ResistInfoData',
    'ResistTypeData',
    'ResourceTypeData',
    'RewardData',
    'RewardTypeData',
    'SkillData',
    'SkillCategoryData',
    'SkillLineData',
    'SkillSetData',
    'TitleData',
    'VehicleAttachmentData',
    'VehicleData',
    'WeaponAmmoSlot',
    'WeaponData',
    'WeaponDatasheet',
    'WorldData',
    'ZoneData',
    'ZoneEffectData',
    'ZoneEffectTypeData'
]



"""Data classes for :mod:`auraxium.ps2.states`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'PlayerStateGroup'
]


class PlayerStateGroup(Ps2Data):
    """A fire-mode-specific state group.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        player_state_group_id: The ID of the player state group.
        player_state_id: The ID of the player state for this entry.
        can_iron_sight: Whether ADS is allowed in this state.
        cof_grow_rate: Cone of fire growth per shot.
        cof_max: The maximum cone-of-fire for this state.
        cof_min: The minimum cone-of-fire for this state.
        cof_recovery_delay_ms: The time before the cone-of-fire will
            start to reset after the user stopped firing.
        cof_recovery_rate: The speed at which the cone-of-fire will
            return to its starting value after the user stopped firing.
        cof_shots_before_penalty: (Not yet documented)
        cof_recovery_delay_threshold: (Not yet documented)
        cof_turn_penalty: (Not yet documented)

    """

    player_state_group_id: int
    player_state_id: int
    can_iron_sight: bool
    cof_grow_rate: Optional[float] = None
    cof_max: float
    cof_min: float
    cof_recovery_delay_ms: Optional[int] = None
    cof_recovery_rate: float
    cof_shots_before_penalty: Optional[int] = None
    cof_recovery_delay_threshold: Optional[int] = None
    cof_turn_penalty: Optional[int] = None



"""Data classes for :mod:`auraxium.ps2.map`."""

from typing import Optional

from ..base import Ps2Data
from ..types import LocaleData

__all__ = [
    'FacilityTypeData',
    'MapHexData',
    'MapRegionData',
    'RegionData'
]


class FacilityTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.map.FacilityType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        facility_type_id: The unique ID of this facility type.
        description: The description of this facility type.

    """

    facility_type_id: int
    description: str


class MapHexData(Ps2Data):
    """Data class for :class:`auraxium.ps2.map.MapHex`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        zone_id: The ID of the zone (or continent) containing this hex.
        map_region_id: The ID of the map region associated with this
            hex.
        x: The X map position of the hex.
        y: The Y map position of the hex.
        hex_type: The type of map hex. Refer to :attr:`type_name` for
            details.
        type_name: The name of the hex' type.

    """

    # pylint: disable=invalid-name

    zone_id: int
    map_region_id: int
    x: int
    y: int
    hex_type: int
    type_name: str


class MapRegionData(Ps2Data):
    """Data class for :class:`auraxium.ps2.map.MapHex`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        map_region_id: The unique ID of this map region.
        zone_id: The ID of the zone (i.e. continent) this region is in.
        facility_id: The ID of the associated facility.
        facility_name: The name of the associated facility.
        facility_type_id: The type ID of the associated facility.
        facility_type: The type name of the associated facility.
        location_x: The X world position of the facility.
        location_y: The Y world position of the facility.
        location_z: The Z world position of the facility.
        reward_amount: (Unused)
        reward_currency_id: (Unused)

    """

    map_region_id: int
    zone_id: int
    facility_id: Optional[int]  # Only missing for the shattered warpgate
    facility_name: str
    facility_type_id: Optional[int]  # Only missing for the shattered warpgate
    facility_type: Optional[str]  # Only missing for the shattered warpgate
    location_x: Optional[float] = None
    location_y: Optional[float] = None
    location_z: Optional[float] = None
    reward_amount: Optional[int] = None
    reward_currency_id: Optional[int] = None


class RegionData(Ps2Data):
    """Data class for :class:`auraxium.ps2.map.Region`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        region_id: The unique ID of the map region.
        zone_id: The ID of the zone (i.e. continent) the region is in.
        initial_faction_id: (Unused)
        name: The localised name of the map region.

    """

    region_id: int
    zone_id: int
    initial_faction_id: int
    name: LocaleData



"""Zone class definition."""

from ..base import Named
from ..models import ZoneData


class Zone(Named, cache_size=20, cache_ttu=3600.0):
    """A continent or dynamic zone in the game world.

    Note that some dynamic zones are not resolvable via this object.
    Dynamic zones include Sanctuary, the Tutorial zones, and the outfit
    wars desolation maps.
    """

    collection = 'zone'
    data: ZoneData
    dataclass = ZoneData
    id_field = 'zone_id'

    # @property
    # def is_dynamic(self) -> bool:
    #     """Return whether the given zone is dynamic or not.

    #     Dynamic zones are spun up for individual players as required,
    #     such as the tutorial world space.
    #     """
    #     # TODO: Add dynamic zone detection



"""Event definitions and utilities for the websocket event stream.

This mostly defines events, lists event types and defines the event
trigger system.

"""

import asyncio
import contextlib
import datetime
import enum
import json
import logging
import warnings
from typing import (Any, Awaitable, Callable, Coroutine, Dict, Iterable,
                    Iterator, List, Optional, Set, TYPE_CHECKING, Union)

import websockets

from .client import Client
from .types import CensusData
from .utils import expo_scaled

if TYPE_CHECKING:
    # This is only imported during static type checking to resolve the forward
    # references. This avoids a circular import at runtime.
    from .ps2 import Character, World

__all__ = [
    'ESS_ENDPOINT',
    'Event',
    'EventClient',
    'EventType',
    'Trigger'
]

# The websocket endpoint to connect to
ESS_ENDPOINT = 'wss://push.planetside2.com/streaming'
log = logging.getLogger('auraxium.ess')

# Pylance is more strict regarding typing of synchronous vs. asynchronous code
Callback = Callable[['Event'], Union[Coroutine[Any, Any, None], None]]


class EventType(enum.IntEnum):
    """A type of event returned by the websocket."""

    # NOTE: IDs 1 through 19 are reserved for character-centric events,
    # IDs 20 through 29 are both character-centric and world-centric, and
    # IDs 30 and up are reserved for world-centric events.

    #: Fall-back value for unknown event types.
    UNKNOWN = 0

    # Character-centric events

    #: A character has earned a new achievement (medal or service ribbon).
    ACHIEVEMENT_EARNED = 1
    #: A character has achieved a new battle rank.
    BATTLE_RANK_UP = 2
    #: A character has been killed.
    DEATH = 3
    #: A character has been granted a new item.
    ITEM_ADDED = 4
    #: A character has been granted a new skill (certification/ASP skill).
    SKILL_ADDED = 5
    #: A character's vehicle has been destroyed.
    VEHICLE_DESTROY = 6
    #: A character has received an experience tick. The field ``other_id`` may
    # may refer to another player, vehicle, or entity.
    # Note that this is a very common event that can break your app if used on
    # large character groups.
    # Use :meth:`EventType.filter_experience()` to filter by experience ID.
    GAIN_EXPERIENCE = 7
    #: A character has participated in the capture of a facility.
    PLAYER_FACILITY_CAPTURE = 8
    #: A character has participated in the defence of a facility.
    PLAYER_FACILITY_DEFEND = 9

    # Character-centric and world-centric events

    #: A character has logged in.
    PLAYER_LOGIN = 20
    #: A character has logged out.
    PLAYER_LOGOUT = 21

    # World-centric events

    #: A continent has been locked.
    CONTINENT_LOCK = 30
    #: A continent has been unlocked. Note that this event is not working as of
    # July 2020.
    CONTINENT_UNLOCK = 31
    #: A facility has changed ownership (i.e. being captured).
    FACILITY_CONTROL = 32
    #: An alert has started or ended.
    METAGAME_EVENT = 33

    @classmethod
    def from_event_name(cls, event_name: str) -> 'EventType':
        """Return the appropriate enum value for the given name.

        The event name is case-insensitive here, but not in general.

        Returns:
            The event type for the given event payload, or
            :attr:`EventType.UNKNOWN`.

        """
        conversion_dict: Dict[str, int] = {
            'AchievementEarned': 1,
            'BattleRankUp': 2,
            'Death': 3,
            'ItemAdded': 4,
            'SkillAdded': 5,
            'VehicleDestroy': 6,
            'GainExperience': 7,
            'PlayerFacilityCapture': 8,
            'PlayerFacilityDefend': 9,
            'PlayerLogin': 20,
            'PlayerLogout': 21,
            'ContinentLock': 30,
            'ContinentUnlock': 31,
            'FacilityControl': 32,
            'MetagameEvent': 33}
        try:
            return cls(conversion_dict[event_name])
        except KeyError:
            # Fall-back for case-insensitive event names
            for name, id_ in conversion_dict.items():
                if name.lower() == event_name.lower():
                    return cls(id_)
            return cls(0)

    @classmethod
    def filter_experience(cls, experience_id: int) -> str:
        """Return a dynamic event matching events by experience ID.

        This works like :attr:`EventType.GAIN_EXPERIENCE`, but allows
        subscribing to only the experience IDs you are interested in.

        This reduces the bandwidth requirements as the number of
        GAIN_EXPERIENCE events can be excessively high for large groups
        of players.

        This returns a string that can be passed to both
        :meth:`EventClient.trigger()` and the :class:`Trigger` class's
        initialiser in place of the enum value itself.

        Arguments:
            experience_id: The experience ID to filter by.

        Returns:
            The event name filtering for the given experience ID.

        """
        event_name = cls.GAIN_EXPERIENCE.to_event_name()
        return f'{event_name}_experience_id_{experience_id}'

    @classmethod
    def from_payload(cls, payload: CensusData) -> 'EventType':
        """Return the appropriate enum value for the event.

        Returns:
            The event type for the given event payload, or
            :attr:`EventType.UNKNOWN`.

        """
        event_name = payload.get('event_name', 'NULL')
        return cls.from_event_name(event_name)

    def to_event_name(self) -> str:
        """Return the event name for the given enum value.

        This is mostly to dynamically subscribe to events.

        Raises:
            ValueError: Raised when calling this method on
                :attr:`EventType.UNKNOWN`.

        Returns:
            The string representing this event.

        """
        conversion_dict: Dict[int, str] = {
            1: 'AchievementEarned',
            2: 'BattleRankUp',
            3: 'Death',
            4: 'ItemAdded',
            5: 'SkillAdded',
            6: 'VehicleDestroy',
            7: 'GainExperience',
            8: 'PlayerFacilityCapture',
            9: 'PlayerFacilityDefend',
            20: 'PlayerLogin',
            21: 'PlayerLogout',
            30: 'ContinentLock',
            31: 'ContinentUnlock',
            32: 'FacilityControl',
            33: 'MetagameEvent'}
        try:
            return conversion_dict[self.value]
        except KeyError as err:
            raise ValueError(
                'Cannot convert EventType.UNKNOWN to event name') from err


class Event:
    """An event returned via the ESS websocket connection.

    The raw response returned through the API is accessible through the
    :attr:`payload` attribute.

    """

    def __init__(self, payload: CensusData) -> None:
        self.timestamp: datetime.datetime = datetime.datetime.utcfromtimestamp(
            int(payload['timestamp']))
        self.payload: CensusData = payload

    @property
    def age(self) -> float:
        """The age of the event in seconds."""
        now = datetime.datetime.now()
        return (self.timestamp - now).total_seconds()

    @property
    def type(self) -> EventType:
        """The type of event."""
        return EventType.from_payload(self.payload)


class Trigger:
    """An event trigger for the client's websocket connection.

    Event triggers encapsulate both the event type to trigger on, as
    well as the action to perform when the event is encountered.

    They are also used to dynamically generate the subscription payload
    required to inform the event streaming service of the event types
    the client wishes to receive.

    Note that some subscriptions are incompatible with each other and
    may require multiple clients to be stable.

    Attributes:
        action: The method or coroutine to run if the matching event is
            encountered.
        characters: A list of characters to filter the incoming events
            by. For some events, like :attr:`EventType.DEATH`, there
            are multiple character IDs involved that may match.
        conditions: Any number of variables or callables that must be
            True for the trigger to run. Note that these filters are
            checked for any matching events, so any callables must be
            synchronous and lightweight.
        events: A set of events that the trigger will listen for.
        last_run: A :class:`datetime.datetime` instance that will be set
            to the last time the trigger has run. This will be
            ``None`` until the first run of the trigger.
        name: The unique name of the trigger.
        single_shot: If True, the trigger will be automatically removed
            from the client when it fires once.
        worlds: A list of worlds to filter the incoming events by.

    """

    def __init__(self, event: Union[EventType, str],
                 *args: Union[EventType, str],
                 characters: Optional[
                     Union[Iterable['Character'], Iterable[int]]] = None,
                 worlds: Optional[
                     Union[Iterable['World'], Iterable[int]]] = None,
                 action: Optional[
                     Callable[[Event], Union[None, Awaitable[None]]]] = None,
                 name: Optional[str] = None,
                 single_shot: bool = False) -> None:
        self.action = action
        self.characters: List[int] = (
            [] if characters is None else [c if isinstance(c, int) else c.id
                                           for c in characters])
        self.conditions: List[Union[bool, Callable[[CensusData], bool]]] = []
        self.events: Set[Union[EventType, str]] = set((event, *args))
        self.last_run: Optional[datetime.datetime] = None
        self.name = name
        self.single_shot = single_shot
        self.worlds: List[int] = (
            [] if worlds is None else [w if isinstance(w, int) else w.id
                                       for w in worlds])

    def callback(self, func: Callable[[Event], None]) -> None:
        """Set the given function as the trigger action.

        The action may be a regular callable or a coroutine.
        Any callable that is a coroutine according to
        :meth:`asyncio.iscoroutinefunction()` will be awaited.

        This method can be used as a decorator.

        .. code-block:: python3

            my_trigger = Trigger('Death')
            @my_trigger.callback
            def pay_respect(event):
                char = event.payload['character_id']
                print('F ({char})')

        Arguments:
            func: The method or coroutine to call when the event
                trigger fires.

        """
        self.action = func

    def check(self, event: Event) -> bool:
        """Return whether the given trigger should fire.

        This only returns whether the trigger should fire, the trigger
        action will be scheduled separately, at which point
        :meth:`Trigger.run()` is called.

        Arguments:
            event: The payload to check.

        Returns:
            Whether this trigger should run for the given event.

        """
        if event.type not in self.events:
            return False
        payload = event.payload
        # Check character ID requirements
        if self.characters:
            char_id = int(payload.get('character_id', 0))
            other_id = int(payload.get('attacker_character_id', 0))
            if char_id not in self.characters or other_id in self.characters:
                return False
        # Check world ID requirements
        if self.worlds:
            if int(payload.get('world_id', 0)) not in self.worlds:
                return False
        # Check custom trigger conditions
        for condition in self.conditions:
            if callable(condition):
                if not condition(payload):
                    return False
            elif not condition:
                return False
        return True

    def generate_subscription(self) -> str:
        """Generate the appropriate subscription for this trigger."""
        json_data: Dict[str, Union[str, List[str]]] = {
            'action': 'subscribe',
            'eventNames': [e.to_event_name() if isinstance(e, EventType) else e
                           for e in self.events],
            'service': 'event'}
        if self.characters:
            json_data['characters'] = [str(c) for c in self.characters]
        else:
            json_data['characters'] = ['all']
        if self.worlds:
            json_data['worlds'] = [str(c) for c in self.worlds]
        else:
            json_data['characters'] = ['all']
        return json.dumps(json_data)

    async def run(self, event: Event) -> None:
        """Perform the action associated with this trigger.

        Arguments:
            event: The event to pass to the trigger action.

        """
        self.last_run = datetime.datetime.now()
        if self.action is None:
            warnings.warn(f'Trigger {self} run with no action specified')
            return
        if asyncio.iscoroutinefunction(self.action):
            await self.action(event)  # type: ignore
        else:
            self.action(event)


class EventClient(Client):
    """Advanced client with event streaming capability.

    This subclass of :class:`Client` extends the interface to also
    provide access to the websocket endpoint at
    ``wss://push.planetside2.com/streaming``.

    To use the websocket endpoint, you have to define a
    :class:`Trigger` and register it using the
    :meth:`EventClient.add_trigger()` method. This will automatically
    open a websocket connection is there is not already one running.
    Likewise, removing all event triggers from the client will cause
    the underlying websocket connection to close.

    Refer to the :class:`Trigger` class's documentation for details on
    how to use triggers and respond to events.

    Attributes:
        triggers: The list of :class:`Triggers <Trigger>` registered
            for the client.
        websocket: The websocket client used for the real-time event
            stream. This will be automatically opened and closed by the
            client as event triggers are added and removed.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.triggers: List[Trigger] = []
        self.websocket: Optional[websockets.WebSocketClientProtocol] = None
        self._connect_lock = asyncio.Lock()
        self._connected: bool = False
        # NOTE: This utility returns a factory, hence the trailing parentheses
        self._reconnect_backoff = self._reset_backoff()
        self._send_queue: List[str] = []

    def add_trigger(self, trigger: Trigger) -> None:
        """Add a new event trigger to the client.

        If there is currently no active websocket connection to the
        event streaming service, one will be created for this trigger.

        .. note::
            As this is a synchronous method, the websocket will not be
            active by the time this method returns.
            Use :meth:`EventClient.wait_ready()` to wait for the
            websocket being ready.

        Arguments:
            trigger: The trigger to add.

        """
        log.debug('Adding trigger %s', trigger)
        self.triggers.append(trigger)
        subscription = trigger.generate_subscription()
        self._send_queue.append(subscription)
        # Only queue the connect() method if it is not already running
        if self.websocket is None and not self._connect_lock.locked():
            log.debug('Websocket not connected, scheduling connection')
            self.loop.create_task(self.connect())

    def find_trigger(self, name: str) -> Trigger:
        """Retrieve a registered event trigger by name.

        If the trigger cannot be found, a :class:`KeyError` is raised.

        Arguments:
            name: [description]

        Raises:
            KeyError: Raised if no trigger with the given name is
                registered for the client.

        """
        for trigger in self.triggers:
            if trigger.name == name:
                return trigger
        raise KeyError(f'No trigger with name "{name}" found')

    def remove_trigger(self, trigger: Union[Trigger, str], *,
                       keep_websocket_alive: bool = False) -> None:
        """Remove an existing event trigger.

        You can either provide the trigger instance to remove, or
        specify the name of the trigger instead.

        By default, the underlying websocket connection will be closed
        if this was the only trigger registered. Use the
        ``keep_websocket_alive`` flag to prevent this.

        Arguments:
            trigger: The trigger to remove, or its unique name.
            keep_websocket_alive (optional): If True, the websocket
                connection will be kept open even if the client has
                zero triggers remaining. Defaults to ``False``.

        Raises:
            KeyError: Raised if a trigger name was passed and no
                trigger of this name exists for the client.
            ValueError: Raised if no trigger of the given name is
                currently registered for this client.

        """
        if not isinstance(trigger, Trigger):
            trigger = self.find_trigger(trigger)
        log.debug('Removing trigger %s', trigger)
        try:
            self.triggers.remove(trigger)
        except ValueError as err:
            raise RuntimeError('The given trigger is not registered for '
                               'this client') from err
        # If this was the only trigger registered, close the websocket
        if not keep_websocket_alive and not self.triggers:
            log.info('All triggers have been removed, closing websocket')
            self.loop.create_task(self.close())

    async def close(self) -> None:
        """Shut down the client.

        This will close the websocket connection and end any ongoing
        HTTP sessions used for requests to the REST API.

        Call this to clean up before the client object is destroyed.

        """
        await self.disconnect()
        await super().close()

    async def connect(self) -> None:
        """Connect to the websocket endpoint and process responses.

        This will continuously loop until :meth:`EventClient.close()`
        is called.
        If the websocket connection encounters and error, it will be
        automatically restarted.

        Add payloads to :attr`EventClient._send_queue` to schedule
        their transmission.

        Any event payloads received will be passed to
        :meth:`EventClient.dispatch()` for filtering and event
        dispatch.

        """
        # NOTE: When multiple triggers are added to the bot without an active
        # websocket connection, this function may be scheduled multiple times.
        if self._connect_lock.locked():
            log.debug('Websocket already running')
            return
        await self._connect_lock.acquire()

        log.info('Connecting to websocket endpoint...')
        url = f'{ESS_ENDPOINT}?environment=ps2&service-id={self.service_id}'
        async with websockets.connect(url) as websocket:
            self.websocket = websocket
            log.info(
                'Connected to %s?environment=ps2&service-id=XXX', ESS_ENDPOINT)
            self._connected = True

            # Reset the backoff generator as the connection attempt was
            # successful
            self._reconnect_backoff = self._reset_backoff()

            # Keep processing websocket events until the connection dies or is
            # closed by the user or trigger system.
            while self._connect_lock.locked():
                try:
                    await self._handle_websocket()
                except websockets.exceptions.ConnectionClosed as err:
                    log.info('Websocket connection closed (%d, %s)',
                             err.code, err.reason)  # type: ignore
                    await self.disconnect()
                    # NOTE: This will increment the reconnect delay each time,
                    # until one connection attempt is successful.
                    delay = next(self._reconnect_backoff)
                    log.info(
                        'Next reconnection attempt in %.2f seconds', delay)
                    await asyncio.sleep(delay)
                    log.info('Attempting to reconnect...')
                    self.loop.create_task(self.connect())

    async def disconnect(self) -> None:
        """Disconnect the websocket.

        Unlike :meth:`EventClient.close()`, this does not affect the
        HTTP session used by regular REST requests.

        """
        if self.websocket is None:
            return
        log.info('Closing websocket connection')
        if self.websocket.open:
            await self.websocket.close()
        with contextlib.suppress(RuntimeError):
            self._connect_lock.release()
        self.websocket = None
        self._connected = False

    def dispatch(self, event: Event) -> None:
        """Dispatch an event to the appropriate event triggers.

        This goes through the list of triggers registered for this
        client and checks if the passed event matches the trigger's
        requirements using :meth:`Trigger.check()`.

        The call-backs for the matching triggers will be scheduled for
        execution in the current event loop using
        :meth:`asyncio.AbstractEventLoop.create_task()`.

        If a trigger's :attr:`Trigger.single_shot` attribute is set to
        True, the trigger will be removed from the client as soon as
        its call-back has been scheduled for execution. This means that
        when the action associated with a single-shot trigger runs, the
        associated trigger will no longer be registered for the client.

        Arguments:
            event: An event received through the event stream.

        """
        # Check for appropriate triggers
        for trigger in self.triggers:
            log.debug('Checking trigger %s', trigger)
            if trigger.check(event):
                log.debug('Scheduling trigger %s', trigger)
                self.loop.create_task(trigger.run(event))
                # Single-shot triggers self-unload as soon as their call-back
                # is scheduled
                if trigger.single_shot:
                    log.info('Removing single-shot trigger %s', trigger)
                    self.remove_trigger(trigger)

    async def _handle_websocket(self, timeout: float = 0.1) -> None:
        """Main loop handling the websocket connection.

        This method processes event payloads, adds automatic reconnect
        capabilities and sends messages added to
        :attr:`EventClient._send_queue`.
        """
        if self.websocket is None:
            return
        try:
            response = str(await asyncio.wait_for(
                self.websocket.recv(), timeout=timeout))
        except asyncio.TimeoutError:
            # NOTE: This inner timeout try block is used to ensure
            # the websocket will regularly check for messages in
            # the client's _send_queue even when no messages are
            # being received.
            # Without this, awaiting self.websocket.recv() would
            # block events from being sent if no responses are
            # received.
            pass
        else:
            log.debug('Received response: %s', response)
            self._process_payload(response)
        finally:
            if self._send_queue:
                msg = self._send_queue.pop(0)
                log.info('Sending message: %s', msg)
                await self.websocket.send(msg)

    def trigger(self, event: Union[str, EventType],
                *args: Union[str, EventType], name: Optional[str] = None,
                **kwargs: Any) -> Callable[[Callback], None]:
        """Create and add a trigger for the given action.

        If no name is specified, the call-back function's name will be
        used as the trigger name.

        Keep in mind that a trigger's name must be unique. A
        :class:`KeyError` will be raised if a trigger with this name
        already exists.

        Arguments:
            event: The event to trigger on.
            *args: Additional events that also trigger the action.
            name (optional): The name to assign to the trigger. If not
                specified, the call-back function's name will be used.
                Defaults to ``None``.

        Raises:
            KeyError: Raised if a trigger with the given name already
                exists.

        """
        trigger = Trigger(event, *args, name=name, **kwargs)

        def wrapper(func: Callback) -> None:
            trigger.action = func
            # If the trigger name has not been specified, use the call-back
            # function's name instead
            if trigger.name is None:
                trigger.name = func.__name__
            # Ensure the trigger name is unique before adding the trigger
            if any(t.name == trigger.name for t in self.triggers):
                raise KeyError(f'The trigger "{trigger.name}" already exists')
            # If the name is unique, register the trigger to the client
            self.add_trigger(trigger)

        return wrapper

    def _process_payload(self, response: str) -> None:
        """Process a response payload received through the websocket.

        This method filters out any non-event messages (such as service
        messages, connection heartbeats or subscription echoes) before
        passing any event payloads on to :meth:`EventClient.dispatch()`.

        Arguments:
            response: The plain text response received through the ESS.

        """
        data: CensusData = json.loads(response)
        service = data.get('service')
        # Event messages
        if service == 'event':
            if data['type'] == 'serviceMessage':
                event = Event(data['payload'])
                log.debug('%s event received, dispatching...', event.type)
                self.dispatch(event)
            elif data['type'] == 'heartbeat':
                log.debug('Heartbeat received: %s', data)
        # Subscription echo
        elif 'subscription' in data:
            log.debug('Subscription echo: %s', data)
        # Service state
        elif data.get('type') == 'serviceStateChange':
            log.info('Service state change: %s', data)
        # Push service
        elif service == 'push':
            log.debug('Ignoring push message: %s', data)
        # Help message
        elif 'send this for help' in data:
            log.info('ESS welcome message: %s', data)
        # Other
        else:
            log.warning('Unhandled message: %s', data)

    @staticmethod
    def _reset_backoff() -> Iterator[float]:
        """Reset the reconnect backoff generator."""
        return expo_scaled(factor=0.1, max_=30.0)()

    async def wait_for(self, trigger: Trigger, *args: Trigger,
                       timeout: Optional[float] = None) -> Event:
        """Wait for one or more triggers to fire.

        This method will wait until any of the given triggers have
        fired, or until the timeout has been exceeded.

        By default, any triggers passed will be automatically removed
        once the first has been triggered, regardless of the triggers'
        :attr:`Trigger.single_shot` setting.

        Arguments:
            trigger: A trigger to wait for.
            *args: Additional triggers that will also resume execution.
            timeout (optional): The maximum number of seconds to wait
                for; never expires if set to ``None``. Defaults to
                ``None``.

        Raises:
            TimeoutError: Raised if the given timeout is exceeded.

        Returns:
            The first event matching the given trigger(s).

        """
        # The following asyncio Event will pause this method until the trigger
        # fires or expires. Think of it as an asynchronous flag.
        async_flag = asyncio.Event()
        # Used to store the event received
        received_event: Optional[Event] = None

        triggers: List[Trigger] = [trigger]
        triggers.extend(args)

        def callback(event: Event) -> None:
            # Store the received event
            nonlocal received_event
            received_event = event
            # Remove the triggers. This suppresses any ValueErrors raised if
            # the trigger that fired was set to single shot mode.
            for trig in triggers:
                # NOTE: Due to Client.dispatch method being a normal method and
                # not a coroutine, it will always remove the trigger itself
                # before this call-back has any chance to fire (even through
                # the call-back itself is synchronous, it is wrapped in the
                # asynchronous Trigger.run method, causing this delay).
                #
                # This means that by the time this code is executed, the
                # trigger will already be removed, meaning that the ValueError
                # will always be raised here and not in Client.dispatch.
                with contextlib.suppress(ValueError):
                    self.remove_trigger(trig)
            # Set the flag to resume execution of the wait_for_event method
            async_flag.set()

        for trig in triggers:
            trig.action = callback
            self.add_trigger(trig)

        # Wait for the triggers to fire, or for the timeout to expire
        if timeout is not None and timeout <= 0.0:
            timeout = None
        try:
            await asyncio.wait_for(async_flag.wait(), timeout=timeout)
        except asyncio.TimeoutError as err:
            raise TimeoutError from err

        assert received_event is not None
        return received_event

    async def wait_ready(self, interval: float = 0.05) -> None:
        """Wait for the websocket connection to be ready.

        This will return once the websocket connection is open and
        active. This condition will be checked regularly as set by the
        ``interval`` argument.

        If the websocket is already active at the time this method is
        called, this will return without delay.

        Arguments:
            interval (optional): The interval at which to check the
                websocket connection's status. Defaults to ``0.05``.

        """
        if self._connected:
            return
        while not self._connected:
            await asyncio.sleep(interval)



"""Object caching system.

This defines a generic cache that can be used to keep local copies of
remote data. This is especially useful for small but expensive queries
like character name resolution.

"""

import dataclasses
import datetime
import logging
import sys
from collections import OrderedDict
from typing import (Any, Dict, Generic, Hashable, Iterable, Iterator, List,
                    Optional, Tuple, TypeVar)

__all__ = [
    'TLRUCache'
]

K = TypeVar('K', bound=Hashable)  # pylint: disable=invalid-name
V = TypeVar('V', bound=Any)  # pylint: disable=invalid-name
log = logging.getLogger('auraxium.cache')


@dataclasses.dataclass()
class CacheItem(Generic[V]):
    """Small dataclass for cache items.

    This can be thought of as a mutable named tuple.

    Attributes:
        value: The instance being cached
        access_counter: The number of times the item has been retrieved
            from the cache.
        first_added: The time the object was added. Used to calculate
            the age of the entry with respect to a :class:`TLRUCache`'s
            :attr`~TLRUCache.ttu` attribute.
        last_accessed: The last time the object was added. Used for the
            least-recently-used component of the cache.

    """

    value: V
    access_counter: int
    first_added: datetime.datetime
    last_accessed: datetime.datetime


class TLRUCache(Generic[K, V]):
    """Basic time-aware Least Recently Used (TLRU) cache.

    This will hold on to a given item for a set amount of time while
    also discarding the least recently used item if the cache is full.

    This uses two parameters: size and ttu (time-to-use). The size
    defines the maximum number of objects the cache may hold at any
    given time. The time-to-use is the number of seconds an object is
    valid in the cache before it expires and must be re-queried.

    Attributes:
        size: The maximum number of items the cache may hold.
        ttu: The time in seconds that a cache item will be valid. If
            set to zero or less, the cache will behave like a regular
            LRU cache.

    """

    def __init__(self, size: int, ttu: float,
                 name: Optional[str] = None) -> None:
        """Initialise a new, empty TLRU cache.

        Arguments:
            size: The maximum number of items in the cache.
            ttu: The time-to-use for items in this cache. Set to zero
                or less to cache objects indefinitely.
            name (optional): A display name to use for this cache.
                Useful for debugging as this name will be used by the
                logs. Defaults to ``None``.

        """
        # NOTE: Mypy currently does not support type hinting the OrderedDict
        # object in-code, hence the string literal type.
        self._data: 'OrderedDict[K, CacheItem[V]]' = OrderedDict()
        self.name = name or 'TLRUCache'
        self.size = size
        self.ttu = ttu

    def __contains__(self, key: str) -> bool:
        """Return whether the given key exists in the cache."""
        return key in self._data.keys()

    def __iter__(self) -> Iterator[K]:
        """Return an iterator over the cache keys."""
        return iter(self._data)

    def __len__(self) -> int:
        """Return the number of elements in the cache.

        This will remove any expired keys before calculation.
        """
        if self.ttu > 0:
            self.remove_expired()
        return len(self._data)

    def add(self, key: K, item: V) -> None:
        """Add a new item to the cache.

        If the cache is full, this will clear any expired items (i.e.
        items who's age is greater than the TTU set for the cache)
        before removing the least recently used item.

        Arguments:
            key: The unique identifier of the object added.
            item: The object to store in the cache.

        """
        now = datetime.datetime.now()
        log.debug('%s: Adding %s instance under key %d',
                  self.name, item.__class__.__name__, key)
        self.free(count=1)
        self._data[key] = CacheItem(item, 0, now, now)

    def add_many(self, items: Iterable[Tuple[K, V]]) -> None:
        """Add multiple items to the cache.

        If the cache is full, this will clear any expired items (i.e.
        items who's age is greater than the TTU set for the cache)
        before removing the necessary number of least recently used
        items.

        Arguments:
            items: An iterable of tuples containing the ID/object pairs
                to add to the cache.

        Raises:
            ValueError: Raised if the number of items to add exceeds
                the size of the cache.

        """
        now = datetime.datetime.now()
        data = {k: CacheItem(v, 0, now, now) for k, v in items}
        if not data:
            log.debug('%s: add_many called with empty iterable', self.name)
            return
        count = len(data)
        if log.isEnabledFor(logging.DEBUG):
            item = next(iter(data.values()))
            log.debug('%s: Adding %d %s instances',
                      self.name, count, item.value.__class__.__name__)
        self.free(count=count)
        self._data.update(data)

    def clear(self) -> None:
        """Clear the cache, removing all items."""
        if log.isEnabledFor(logging.DEBUG):
            count = len(self._data)
            log.debug('%s: Clearing cache: %d item%s will be removed',
                      self.name, count, 's' if count > 1 else '')
        self._data.clear()

    def footprint(self) -> int:
        """Return the size of the cache items in bytes.

        Note that this is not an exact value, but can still be useful
        for profiling.

        Returns:
            The size of the cache in bytes.

        """
        return sys.getsizeof(self._data)

    def free(self, count: int = 1) -> int:
        """Release items from the cache.

        This will clear any expired items, followed by as many of the
        least recently used items as required to accommodate the number
        of items specified via the ``count`` argument.

        The number of slots freed may exceed the number of slots
        requested due to all expired items being cleared.

        Arguments:
            count (optional): The number of free slots to request.
                Defaults to ``1``.

        Raises:
            ValueError: Raised if the number of requested free slots
                exceeds the size of the cache.

        Returns:
            The number of available slots in the cache. This may be
            greater than the number of slots requested.

        """
        if count > self.size:
            raise ValueError(f'Unable to provide {count} available slots, '
                             f'{self.name} can only hold {self.size} items')
        available = self.size - len(self._data)
        if log.isEnabledFor(logging.DEBUG):
            log.debug('%s: %d slot%s requested (available: %d)',
                      self.name, count, 's' if count > 1 else '', available)
        if self.ttu > 0:
            available += self.remove_expired()
        if available >= count:
            return available
        self.remove_lru(count=count-available)
        return count

    def get(self, key: K) -> Optional[V]:
        """Retrieve an item from the cache.

        Arguments:
            key: The unique identifier of the object to retrieve.

        Returns:
            The object stored under the given identifier, or ``None``
            if it is not found or expired (i.e. its age exceeds the TTU
            set for the cache).

        """
        now = datetime.datetime.now()
        try:
            item = self._data[key]
        except KeyError:
            log.debug('%s: Key %s not found', self.name, key)
            return None
        item.access_counter += 1
        if self.ttu > 0:
            age = now - item.first_added
            if age.total_seconds() > self.ttu:
                log.debug(
                    '%s: Key %d expired, age: %.1f sec. (max: %.1f sec.)',
                    self.name, key, age.total_seconds(), self.ttu)
                del self._data[key]
                return None
        else:
            log.debug('%s: Skipping expiration check (TTU %d)',
                      self.name, self.ttu)
        log.debug('%s: Key %d found, moving to top', self.name, key)
        self._data.move_to_end(key, last=True)
        item.last_accessed = now
        self._data[key] = item
        return item.value

    def items(self) -> Dict[K, V]:
        """Return a mapping of all key/value pairs in the cache.

        Note that this is mostly intended for introspection and
        troubleshooting, you should only use :meth:`TLRUCache.get()` to
        retrieve items from the cache.

        This method will not update the items'
        :attr:`CacheItem.last_accessed` value or increment their
        :attr:`CacheItem.access_counter`.

        Returns:
            A dictionary containing all items in the cache, with more
            recently accessed items first.

        """
        return {k: v.value for k, v in self._data.items()}

    def last_accessed(self, key: K) -> datetime.datetime:
        """Return the time the given item was last accessed.

        Unlike :meth:`TLRUCache.get()`, this does not perform an
        expiration check and also will not push the retrieved item to
        the top of the cache.

        This method is intended for introspection and troubleshooting,
        but is also not very expensive.

        Arguments:
            key: The unique identifier of the item to check.

        Raises:
            ValueError: Raised if the given identifier is not found.

        Returns:
            The time the given item was last accessed.

        """
        try:
            item = self._data[key]
        except KeyError as err:
            raise ValueError(f'Key not found: {key}') from err
        return item.last_accessed

    def remove_expired(self) -> int:
        """Remove any expired items from the cache.

        Returns:
            The number of items removed from the cache.

        """
        if self.ttu <= 0:
            logging.warning('%s: remove_expired called with TTU disabled',
                            self.name)
            return 0
        now = datetime.datetime.now()
        keys_to_remove: List[K] = []
        for key, data in self._data.items():
            age = now - data.first_added
            if age.total_seconds() > self.ttu:
                keys_to_remove.append(key)
        _ = [self._data.pop(k) for k in keys_to_remove]
        count = len(keys_to_remove)
        log.debug('%s: Removed %d expired items', self.name, count)
        return count

    def remove_lru(self, count: int = 1) -> None:
        """Remove the given number of items from the cache.

        This will remove the least recently used (LRU) items first.

        Arguments:
            count (optional): The number of items to remove. Defaults
            to ``1``.

        Raises:
            ValueError: raised if the number of items to remove is
                negative
            ValueError: Raised if the number of items to remove exceeds
                the size of the cache.

        """
        if count < 0:
            raise ValueError('count may not be negative')
        if count > self.size:
            raise ValueError(f'Unable to remove {count} items from '
                             f'{self.name}, cache size is set to {self.size} '
                             'items')
        log.debug('%s: Removing %d LRU items', self.name, count)
        for _ in range(count):
            _ = self._data.popitem(last=True)

    def values(self) -> List[V]:
        """Return a list of all items in the cache.

        Note that this is mostly intended for introspection and
        troubleshooting, you should only use TLRUCache.get() to
        retrieve items from the cache.

        Returns:
            A list of all items in the cache.

        """
        return [v.value for v in self._data.values()]



"""Data classes for :mod:`auraxium.ps2.objective`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'ObjectiveData',
    'ObjectiveTypeData'
]


class ObjectiveData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.Objective`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        objective_id: The unique ID of this objective.
        objective_type_id: The associated objective type for this
            objective.
        objective_group_id: The objective group this objective
            contributes to. Used to link objectives to directives.
        param*: Type-specific parameters for this objective. Refer to
            the corresponding :class:`ObjectiveType` for details.

    """

    objective_id: int
    objective_type_id: int
    objective_group_id: int
    param1: Optional[str] = None
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None
    param6: Optional[str] = None
    param7: Optional[str] = None
    param8: Optional[str] = None
    param9: Optional[str] = None


class ObjectiveTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.ObjectiveType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        objective_type_id: The unique ID of the objective type.
        description: A description of what the objective type is used
            for.
        param*: Descriptions of what the corresponding parameter is
            used for in objectives of this type.

    """

    objective_type_id: int
    description: str
    param1: Optional[str] = None
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None
    param6: Optional[str] = None
    param7: Optional[str] = None
    param8: Optional[str] = None
    param9: Optional[str] = None



"""Reward and reward type class definitions."""

from typing import Final

from ..base import Cached
from ..census import Query
from ..client import Client
from ..models import RewardData, RewardTypeData
from ..proxy import InstanceProxy, SequenceProxy


class RewardType(Cached, cache_size=10, cache_ttu=3600.0):
    """A type of reward.

    This class mostly specifies the purpose of any generic parameters.
    """

    collection = 'reward_type'
    data: RewardTypeData
    dataclass = RewardTypeData
    id_field = 'reward_type_id'


class Reward(Cached, cache_size=50, cache_ttu=60.0):
    """A reward awarded to a player.

    Access the corresponding :class:`auraxium.ps2.reward.RewardType`
    instance via the :meth:`type` method for information on generic
    parameters.
    """

    collection = 'reward'
    data: RewardData
    dataclass = RewardData
    id_field = 'reward_id'

    @classmethod
    def get_by_reward_group(cls, reward_group_id: int, client: Client
                            ) -> SequenceProxy['Reward']:
        """Return any rewards contained in the given reward group."""
        collection: Final[str] = 'reward_group_to_reward'
        query = Query(collection, service_id=client.service_id)
        query.add_term(field='reward_group_id', value=reward_group_id)
        query.limit(100)
        join = query.create_join(Reward.collection)
        join.set_fields(Reward.id_field)
        return SequenceProxy(Reward, query, client=client)

    @classmethod
    def get_by_reward_set(cls, reward_set_id: int, client: Client
                          ) -> SequenceProxy['Reward']:
        """Return any rewards contained in the given reward set."""
        collection: Final[str] = 'reward_set_to_reward_group'
        query = Query(collection, service_id=client.service_id)
        query.add_term(field='reward_set_id', value=reward_set_id)
        query.limit(100)
        join = query.create_join('reward_group_to_reward').set_list(True)
        join.set_fields('reward_group_id')
        nested = join.create_join(Reward.collection)
        nested.set_fields(Reward.id_field)
        return SequenceProxy(Reward, query, client=client)

    def type(self) -> InstanceProxy[RewardType]:
        """Return the type of reward."""
        query = Query(
            RewardType.collection, service_id=self._client.service_id)
        query.add_term(
            field=RewardType.id_field, value=self.data.reward_type_id)
        return InstanceProxy(RewardType, query, client=self._client)



"""Define the proxy object system."""

import asyncio
import datetime
import warnings
from typing import (Any, Dict, Generator, Generic, Iterator, List, Optional,
                    Type, TypeVar)

from .base import Ps2Object
from .census import JoinedQuery, Query
from .client import Client
from .request import extract_payload

__all__ = [
    'InstanceProxy',
    'Proxy',
    'SequenceProxy'
]

Ps2ObjectT = TypeVar('Ps2ObjectT', bound=Ps2Object)


class Proxy(Generic[Ps2ObjectT]):
    """Base class for any proxy objects.

    The query object passed must specify the parent field name for all
    of its joins. This is necessary to allow parsing of the payload.

    Additionally, this currently does not support custom insertion
    fields.

    Attributes:
        query: The API query used to populate the proxy object.

    """

    def __init__(self, type_: Type[Ps2ObjectT], query: Query,
                 client: Client, lifetime: float = 60.0) -> None:
        """Initialise the proxy.

        Note that the lifetime argument may not exceed the UTC epoch
        seconds due to the way this value is initialised.

        Arguments:
            type_: The object type represented by the proxy.
            query: The query used to retrieve the data.
            client: The client through which to query the API.
            lifetime (optional): The time-to-use of the retrieved data.
                Defaults to ``60.0``.

        """
        self._type = type_
        self.query = query
        self._client = client
        self._ttu = lifetime
        self._data: List[Ps2ObjectT]
        self._index: int
        self._lock = asyncio.Lock()
        self._last_fetched = datetime.datetime.utcfromtimestamp(0)
        max_age = datetime.datetime.now() - self._last_fetched
        assert self._ttu < max_age.total_seconds()

    async def _poll(self) -> None:
        """Query the API, retrieving the data.

        This method uses a lock to ensure it does not try to query the
        same object multiple times.

        """
        async with self._lock:
            payload = await self._client.request(self.query)
            list_ = self._resolve_nested_payload(payload)
            # NOTE: There does not appear to be an easy way to type something
            # as "subclass of an abstract class and all abstract methods have
            # been overwritten", which is why this is typed as being a
            # Ps2Object subclass and then promptly ignored here.
            self._data = [self._type(  # type: ignore
                data, client=self._client) for data in list_]
            self._last_fetched = datetime.datetime.now()

    def _resolve_nested_payload(self, payload: Dict[str, Any]
                                ) -> List[Dict[str, Any]]:
        """Resolve the object payload.

        This introspects the given query to determine the sub-key for
        the actual object to return.

        Arguments:
            payload: The raw payload returned from the API.

        Raises:
            RuntimeError: Raised if the query has more than one join.
            RuntimeError: Raised if the parent field of a query is not
                given.

        Returns:
            The native list of payloads, ready for instantiation.

        """

        def resolve_join(join: JoinedQuery, parent: List[Dict[str, Any]]
                         ) -> List[Dict[str, Any]]:
            # NOTE: The parent list will always be a list of all items the join
            # has been applied to. The resulting list should be a merged set of
            # the elements therein.
            assert join.data.collection is not None
            if (on_ := join.data.field_on) is None:
                raise RuntimeError('Parent field of None not supported')
            # Filter out the payload sub-dict corresponding to the given join
            data: List[Dict[str, Any]] = []
            for element in parent:
                value = element[f'{on_}_join_{join.data.collection}']
                if join.data.is_list:
                    data.extend(value)
                else:
                    data.append(value)
            # Recursively resolve any inner joins
            if join.joins:
                parent = data
                data = []
                for inner in join.joins:
                    data.extend(resolve_join(inner, parent))
            return data

        # Main query
        assert self.query.data.collection is not None
        data = extract_payload(payload, self.query.data.collection)
        # Resolve any joins
        if self.query.joins:
            parent = data
            # If any joins were defined, resolve each of the joins and merge
            # their outputs before returning
            data = []
            for join in self.query.joins:
                data.extend(resolve_join(join, parent))
        return data


class SequenceProxy(Proxy[Ps2ObjectT]):
    """Proxy for lists of results.

    Use this is your joins are returning a list of objects.

    """

    def __aiter__(self) -> 'SequenceProxy[Ps2ObjectT]':
        self._index = -1
        return self

    async def __anext__(self) -> Ps2ObjectT:
        age = datetime.datetime.now() - self._last_fetched
        if age.total_seconds() > self._ttu:
            if self._index > -1:
                warnings.warn('Data went stale during iteration, polling new')
            await self._poll()
        self._index += 1
        try:
            return self._data[self._index]
        except IndexError as err:
            raise StopAsyncIteration from err

    def __await__(self) -> Iterator[List[Ps2ObjectT]]:
        return self.flatten().__await__()

    async def flatten(self) -> List[Ps2ObjectT]:
        """Retrieve all elements in the response as a list.

        Returns:
            A list of instantiated objects.

        """
        return [e async for e in self]


class InstanceProxy(Proxy[Ps2ObjectT]):
    """Proxy for lists of results.

    Use this is your joins are returning a single object.

    """

    def __await__(self) -> Generator[Any, None, Optional[Ps2ObjectT]]:
        return self.resolve().__await__()

    async def resolve(self) -> Optional[Ps2ObjectT]:
        """Return the proxy object.

        Returns:
            The object, or None if no match was found.

        """
        age = datetime.datetime.now() - self._last_fetched
        if age.total_seconds() > self._ttu:
            await self._poll()
        try:
            return self._data[0]
        except IndexError:
            return None



"""Objective class definitions."""

from ..base import Cached
from ..census import Query
from ..models import ObjectiveData, ObjectiveTypeData
from ..proxy import InstanceProxy


class ObjectiveType(Cached, cache_size=10, cache_ttu=60.0):
    """A type of objective.

    This class mostly specifies the purpose of any generic parameters.
    """

    collection = 'objective_type'
    data: ObjectiveTypeData
    dataclass = ObjectiveTypeData
    id_field = 'objective_type_id'


class Objective(Cached, cache_size=10, cache_ttu=60.0):
    """A objective presented to a character."""

    collection = 'objective'
    data: ObjectiveData
    dataclass = ObjectiveData
    id_field = 'objective_id'

    def type(self) -> InstanceProxy[ObjectiveType]:
        """Return the objective type of this objective."""
        query = Query(
            ObjectiveType.collection, service_id=self._client.service_id)
        query.add_term(
            field=ObjectiveType.id_field, value=self.data.objective_type_id)
        return InstanceProxy(ObjectiveType, query, client=self._client)



"""Data classes for :mod:`auraxium.ps2.reward`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'RewardData',
    'RewardTypeData'
]


class RewardData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.Reward`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        reward_id: The unique ID of this reward.
        reward_type_id: The :class:`RewardType` of this reward.
        count_min: The minimum number of rewarded items/currency.
        count_max: The maximum number of rewarded items/currency.
        param*: Type-specific parameters for this reward. Refer to the
            corresponding :class:`RewardType` for details.

    """

    reward_id: int
    reward_type_id: int
    count_min: int
    count_max: int
    param1: str
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None


class RewardTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.ability.ResourceType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        reward_type_id: The unique ID of this reward type.
        description: A description of what this reward type is used
            for.
        count_min: The minimum number of rewarded items/currency.
        count_max: The maximum number of rewarded items/currency.
        param*: Descriptions of what the corresponding parameter is
            used for in rewards of this type.

    """

    reward_type_id: int
    description: str
    count_min: Optional[str]
    count_max: Optional[str]
    param1: str
    param2: Optional[str] = None
    param3: Optional[str] = None
    param4: Optional[str] = None
    param5: Optional[str] = None



"""Data classes for :mod:`auraxium.ps2.faction`."""

from ..base import ImageData, Ps2Data
from ..types import LocaleData

__all__ = [
    'FactionData'
]


class FactionData(Ps2Data, ImageData):
    """Data class for :class:`auraxium.ps2.faction.Faction`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        faction_id: The unique ID of this faction.
        name: The localised name of the faction.
        code_tag: The canonical tag representation of the faction.
        user_selectable: Whether this faction is playable.

    """

    faction_id: int
    name: LocaleData
    code_tag: str
    user_selectable: bool



"""Data classes for :mod:`auraxium.ps2.resist`."""

from typing import Optional

from ..base import Ps2Data

__all__ = [
    'ResistInfoData',
    'ResistTypeData'
]


class ResistInfoData(Ps2Data):
    """Data class for :class:`auraxium.ps2.armour.ResistInfo`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        resist_info: The ID of this resist info entry.
        resist_type_id: The ID of the :class:`ResistType` for this
            entry.
        resist_percent: The damage reduction in percent.
        resist_amount: A flat amount of damage to absorb.
        multiplier_when_headshot: A headshot multiplier override to
            apply.
        description: A description of this resist info entry.

    """

    resist_info_id: int
    resist_type_id: int
    resist_percent: Optional[int] = None
    resist_amount: Optional[int] = None
    multiplier_when_headshot: Optional[float] = None
    description: str


class ResistTypeData(Ps2Data):
    """Data class for :class:`auraxium.ps2.armour.ResistType`.

    This class mirrors the payload data returned by the API, you may
    use its attributes as keys in filters or queries.

    Attributes:
        resist_type_id: The unique ID of this resist type.
        description: A description of what this resist type is used
            for.

    """

    resist_type_id: int
    description: str



"""Effect and effect type class definitions."""

import enum
from typing import Optional

from ..base import Cached
from ..census import Query
from ..models import EffectData, EffectTypeData
from ..proxy import InstanceProxy


class TargetType(enum.IntEnum):
    """Enumerate the types of targets for effects."""

    SELF = 1
    ANY = 2
    ENEMY = 3
    ALLY = 4


class EffectType(Cached, cache_size=20, cache_ttu=60.0):
    """A type of effect.

    This class mostly specifies the purpose of any generic parameters.
    """

    collection = 'effect_type'
    data: EffectTypeData
    dataclass = EffectTypeData
    id_field = 'effect_type_id'


class Effect(Cached, cache_size=10, cache_ttu=60.0):
    """An effect acting on a character.

    Access the corresponding :class:`auraxium.ps2.effect.EffectType`
    instance via the :meth:`type` method for information on generic
    parameters.
    """

    collection = 'effect'
    data: EffectData
    dataclass = EffectData
    id_field = 'effect_id'

    def target_type(self) -> Optional[TargetType]:
        """Return the target type of this effect."""
        if self.data.target_type_id is None:
            return None
        return TargetType(self.data.target_type_id)

    def type(self) -> InstanceProxy[EffectType]:
        """Return the effect type of this effect."""
        query = Query(
            EffectType.collection, service_id=self._client.service_id)
        query.add_term(
            field=EffectType.id_field, value=self._client.service_id)
        return InstanceProxy(EffectType, query, client=self._client)
